from __future__ import annotations

__all__ = ["Connection", "HydraulicLoop"]

import dataclasses as _dc
import typing as _tp

from trnsysGUI.hydraulicLoops import _serialization as _ser

if _tp.TYPE_CHECKING:
    import trnsysGUI.connection.singlePipeConnection as _spc


@_dc.dataclass
class Connection:
    name: str
    diameterInCm: float
    uValueInWPerM2K: float
    lengthInM: float
    connection: _spc.SinglePipeConnection


@_dc.dataclass
class HydraulicLoop:
    name: str
    fluid: _ser.Fluid
    useLoopWideDefaults: bool
    connections: _tp.Sequence[Connection]



import typing as _tp
import pathlib as _pl

import pytrnsys.utils.result as _res
from trnsysGUI.BlockItem import BlockItem


def getPlaceholderValues(ddckDirNames: _tp.Sequence[str], trnsysObjects) -> _res.Result[dict]:
    ddckPlaceHolderValuesDictionary = {}
    for component in trnsysObjects:
        if not isinstance(component, BlockItem) or not component.hasDdckPlaceHolders():
            continue

        if not component.path:
            return _res.Error(
                f"{component.displayName} doesn't have ddck template path"
            )

        componentPath = _pl.Path(component.path)
        componentName = componentPath.stem
        if componentName not in ddckDirNames:
            return _res.Error(
                f"{componentName} is not a ddck template path"
            )

        inputDct = {}
        for inputPort in component.inputs:
            internalPiping = inputPort.parent.getInternalPiping()
            portItemsInternalRealNode = internalPiping.getPortItemsAndAdjacentRealNodeForGraphicalPortItem(
                inputPort)
            portItems = [pr.portItem for pr in portItemsInternalRealNode]
            formattedPortItems = [f"{p.name}" for p in portItems]
            inputDct[formattedPortItems[0]] = {
                "@temp": "T" + inputPort.connectionList[0].displayName,
                "@mfr": "Mfr" + inputPort.connectionList[0].displayName}

        if not inputDct:
            return _res.Error(
                f"{componentName} doesn't have connection name for placeholder"
            )
        ddckPlaceHolderValuesDictionary[componentName] = inputDct

    return ddckPlaceHolderValuesDictionary



import logging as _log
import pathlib as _pl

import PyQt5.QtWidgets as _qtw
import pytest as _pt

import pytrnsys.utils.result as _res
import trnsysGUI.diagram.Editor as _de

_DATA_DIR_ = _pl.Path(__file__).parent / "data"


class TestAutomaticConnection:
    def testConnectionJson(self, request: _pt.FixtureRequest):
        actualDirPath = _DATA_DIR_ / "TRIHP_dualSource"
        expectedDirPath = _DATA_DIR_ / "expected"
        actualJsonFilePath = actualDirPath / "DdckPlaceHolderValues.json"
        expectedJsonFilePath = expectedDirPath / "DdckPlaceHolderValues.json"

        # The following line is required otherwise QT will crash
        application = _qtw.QApplication([])

        def quitApplication():
            application.quit()

        request.addfinalizer(quitApplication)

        editor = self._createEditor(actualDirPath)
        result = editor.encodeDdckPlaceHolderValuesToJson(actualJsonFilePath)
        assert not _res.isError(result)

        actualJsonText = actualJsonFilePath.read_text()  # pylint: disable=bad-option-value,unspecified-encoding
        expectedJsonText = expectedJsonFilePath.read_text()  # pylint: disable=bad-option-value,unspecified-encoding

        assert actualJsonText == expectedJsonText

    @staticmethod
    def _createEditor(projectFolderPath):  # pylint: disable=duplicate-code
        logger = _log.Logger("root")
        editor = _de.Editor(
            parent=None,
            projectFolder=str(projectFolderPath),
            jsonPath=None,
            loadValue="load",
            logger=logger,
        )
        return editor



import typing as _tp

import trnsysGUI.PortItemBase as _pi
import trnsysGUI.doublePipeModelPortItems as _dpmpi
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn


class DoublePipePortItem(_pi.PortItemBase):
    def _selectConnectedRealNode(  # pylint: disable=duplicate-code  # 1
            self,
            portItem: _mfn.PortItem,
            connectedPortItemsAndAdjacentRealNode: _tp.Sequence[_mfs.PortItemAndAdjacentRealNode],
    ) -> _mfn.RealNodeBase:
        assert isinstance(
            portItem, (_dpmpi.ColdPortItem, _dpmpi.HotPortItem)
        ), "Only hot or cold model port items should be mapped to double pipe graphical port items."

        selectedRealNodes = [
            pan.realNode
            for pan in connectedPortItemsAndAdjacentRealNode
            if type(pan.portItem) == type(portItem)  # pylint: disable=unidiomatic-typecheck
        ]

        assert (
                len(selectedRealNodes) == 1
        ), "Only exactly one hot (or cold) model port item should be connected to a double pipe graphical port item."

        selectedRealNode = selectedRealNodes[0]

        return selectedRealNode

    def _highlightInternallyConnectedPortItems(self):
        pass

    def _unhighlightInternallyConnectedPortItems(self):
        pass



import typing as _tp

from PyQt5 import QtGui, QtCore
from PyQt5.QtCore import QLineF
from PyQt5.QtWidgets import QGraphicsLineItem

from trnsysGUI.SegmentItemBase import SegmentItemBase  # type: ignore[attr-defined]

# This is needed to avoid a circular import but still be able to type check
if _tp.TYPE_CHECKING:
    from trnsysGUI.connection.connectionBase import ConnectionBase  # type: ignore[attr-defined]  #  pylint: disable=unused-import


class DoublePipeSegmentItem(SegmentItemBase):
    def __init__(self, startNode, endNode, parent: "ConnectionBase"):
        super().__init__(startNode, endNode, parent)

        self.blueLine = QGraphicsLineItem(self)
        self.redLine = QGraphicsLineItem(self)

    def _createSegment(self, startNode, endNode) -> "SegmentItemBase":
        return DoublePipeSegmentItem(startNode, endNode, self.connection)

    def _setLineImpl(self, x1, y1, x2, y2):
        self.blueLine.setPen(QtGui.QPen(QtCore.Qt.blue, 3))
        self.redLine.setPen(QtGui.QPen(QtCore.Qt.red, 3))
        offset = 3

        if abs(y1 - y2) < 1:
            self.blueLine.setLine(x1, y1 + offset, x2, y2 + offset)
            self.redLine.setLine(x1, y1 - offset, x2, y2 - offset)
        elif abs(x1 - x2) < 1:
            self.blueLine.setLine(x1 + offset, y1, x2 + offset, y2)
            self.redLine.setLine(x1 - offset, y1, x2 - offset, y2)
        else:
            # Initially, connections go directly between its two ports: only
            # afterwards are they broken up into horizontal and vertical segments.
            # This is probably an artifact of back when connections didn't have to
            # be straight. Once, these artifacts have been removed, we should throw here.
            pass
        self.linePoints = QLineF(x1, y1, x2, y2)

    def updateGrad(self):
        self.blueLine.setPen(QtGui.QPen(QtCore.Qt.blue, 3))
        self.redLine.setPen(QtGui.QPen(QtCore.Qt.red, 3))

    def setSelect(self, isSelected: bool) -> None:
        if isSelected:
            selectPen = self._createSelectPen()
            self.blueLine.setPen(selectPen)
            self.redLine.setPen(selectPen)
        else:
            self.updateGrad()

    def setColorAndWidthAccordingToMassflow(self, color, width):
        pass



# pylint: skip-file
# type: ignore

import os
import shutil

from PyQt5.QtCore import QSize
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QTreeView

from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.MyQFileSystemModel import MyQFileSystemModel
from trnsysGUI.MyQTreeView import MyQTreeView


class Control(BlockItem):
    def __init__(self, trnsysType, parent, **kwargs):
        super(Control, self).__init__(trnsysType, parent, **kwargs)
        factor = 0.667  # 0.63 for png
        self.w = 100
        self.h = 100
        self.loadedFiles = []

        self.addTree()

    def addTree(self):
        """
        When a blockitem is added to the main window.
        A file explorer for that item is added to the right of the main window by calling this method
        """
        print(self.parent.parent())
        pathName = self.displayName
        if self.parent.parent().projectPath == "":
            # self.path = os.path.dirname(__file__)
            # self.path = os.path.join(self.path, 'default')
            self.path = self.parent.parent().projectFolder
            # now = datetime.now()
            # self.fileName = now.strftime("%Y%m%d%H%M%S")
            # self.path = os.path.join(self.path, self.fileName)
        else:
            self.path = self.parent.parent().projectPath
        self.path = os.path.join(self.path, "ddck")
        self.path = os.path.join(self.path, pathName)
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.model = MyQFileSystemModel()
        self.model.setRootPath(self.path)
        self.model.setName(self.displayName)
        self.tree = MyQTreeView(self.model, self)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(self.path))
        self.tree.setObjectName("%sTree" % self.displayName)
        for i in range(1, self.model.columnCount() - 1):
            self.tree.hideColumn(i)
        self.tree.setMinimumHeight(200)
        self.tree.setSortingEnabled(True)
        self.parent.parent().splitter.addWidget(self.tree)

    def deleteBlock(self):
        """
        Overridden method to also delete folder
        """
        print("Block " + str(self) + " is deleting itself (" + self.displayName + ")")
        self.deleteConns()
        # print("self.parent.parent" + str(self.parent.parent()))
        self.parent.parent().trnsysObj.remove(self)
        print("deleting block " + str(self) + self.displayName)
        # print("self.scene is" + str(self.parent.scene()))
        self.parent.scene().removeItem(self)
        widgetToRemove = self.parent.parent().findChild(QTreeView, self.displayName + "Tree")
        shutil.rmtree(self.path)
        self.deleteLoadedFile()
        try:
            widgetToRemove.hide()
        except AttributeError:
            print("Widget doesnt exist!")
        else:
            print("Deleted widget")
        del self

    def setDisplayName(self, newName):
        """
        Overridden method to also change folder name
        """
        self.displayName = newName
        self.label.setPlainText(newName)
        self.model.setName(self.displayName)
        self.tree.setObjectName("%sTree" % self.displayName)
        print(os.path.dirname(self.path))
        destPath = os.path.join(os.path.split(self.path)[0], self.displayName)
        if os.path.exists(self.path):
            os.rename(self.path, destPath)
            self.path = destPath
            print(self.path)



import trnsysGUI.massFlowSolver.networkModel as _mfn


class ColdPortItem(_mfn.PortItem):
    def canOverlapWith(self, other: _mfn.PortItem):
        isCompatibleType = type(other) == _mfn.PortItem or isinstance(  # pylint: disable=unidiomatic-typecheck
            other, ColdPortItem
        )
        return super().canOverlapWith(other) and isCompatibleType


class HotPortItem(_mfn.PortItem):
    def canOverlapWith(self, other: _mfn.PortItem):
        isCompatibleType = type(other) == _mfn.PortItem or isinstance(  # pylint: disable=unidiomatic-typecheck
            other, HotPortItem
        )
        return super().canOverlapWith(other) and isCompatibleType



import typing as _tp

import trnsysGUI.images as _img
from trnsysGUI.sourceSinkBase import SourceSinkBase


class Sink(SourceSinkBase):
    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.SINK_SVG

    def changeSize(self):
        self._positionLabel()

        self.origInputsPos = [[40, 0]]
        self.origOutputsPos = [[20, 0]]

        # pylint: disable=duplicate-code  # 1
        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 1 + 2 * self.flippedV) % 4
        self.outputs[0].side = (self.rotationN + 1 + 2 * self.flippedV) % 4
        # pylint: disable=duplicate-code  # 1



from __future__ import annotations

import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import PyQt5.QtWidgets as _qtw
import dataclasses_jsonschema as _dcj

import pytrnsys.utils.serialization as _ser
import trnsysGUI.doublePipePortItem as _dppi
import trnsysGUI.massFlowSolver.networkModel as _mfn
import trnsysGUI.connection.connectionBase as _cb
import trnsysGUI.doublePipeModelPortItems as _dpmpi
import trnsysGUI.doublePipeSegmentItem as _dpsi
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.connectorsAndPipesExportHelpers as _helpers


class DoublePipeConnection(_cb.ConnectionBase):
    def __init__(self, fromPort: _dppi.DoublePipePortItem, toPort: _dppi.DoublePipePortItem, parent):
        super().__init__(fromPort, toPort, parent)

        self.childIds = []
        self.childIds.append(self.trnsysId)
        self.childIds.append(self.parent.idGen.getTrnsysID())

        self._updateModelPipes(self.displayName)

    def setDisplayName(self, newName: str) -> None:
        super().setDisplayName(newName)
        self._updateModelPipes(newName)

    @property
    def fromPort(self) -> _dppi.DoublePipePortItem:
        assert isinstance(self._fromPort, _dppi.DoublePipePortItem)
        return self._fromPort

    @property
    def toPort(self) -> _dppi.DoublePipePortItem:
        assert isinstance(self._toPort, _dppi.DoublePipePortItem)
        return self._toPort

    def _createSegmentItem(self, startNode, endNode):
        return _dpsi.DoublePipeSegmentItem(startNode, endNode, self)

    def getRadius(self):
        rad = 4
        return rad

    def createDeleteUndoCommand(self, parentCommand: _tp.Optional[_qtw.QUndoCommand] = None) -> _qtw.QUndoCommand:
        undoCommand = DeleteDoublePipeConnectionCommand(self, parentCommand)
        return undoCommand

    def encode(self):
        if len(self.segments) > 0:
            labelPos = self.segments[0].label.pos().x(), self.segments[0].label.pos().y()
            labelMassPos = self.segments[0].labelMass.pos().x(), self.segments[0].labelMass.pos().y()
        else:
            self.logger.debug("This connection has no segment")
            defaultPos = self.fromPort.pos().x(), self.fromPort.pos().y()  # pylint: disable = duplicate-code # 1
            labelPos = defaultPos
            labelMassPos = defaultPos

        corners = []
        for corner in self.getCorners():
            cornerTupel = (corner.pos().x(), corner.pos().y())
            corners.append(cornerTupel)

        doublePipeConnectionModel = DoublePipeConnectionModel(
            self.connId,
            self.displayName,
            self.id,
            self.childIds,
            corners,
            labelPos,
            labelMassPos,
            self.fromPort.id,
            self.toPort.id,
            self.trnsysId,
        )

        dictName = "Connection-"
        return dictName, doublePipeConnectionModel.to_dict()

    def decode(self, i):
        model = DoublePipeConnectionModel.from_dict(i)

        self.id = model.id
        self.connId = model.connectionId
        self.trnsysId = model.trnsysId
        self.childIds = model.childIds
        self.setDisplayName(model.name)

        if len(model.segmentsCorners) > 0:
            self.loadSegments(model.segmentsCorners)

        self.setLabelPos(model.labelPos)
        self.setMassLabelPos(model.massFlowLabelPos)

    def getInternalPiping(self) -> _mfs.InternalPiping:
        coldModelPortItemsToGraphicalPortItem = {
            self._coldPipe.fromPort: self.toPort,
            self._coldPipe.toPort: self.fromPort,
        }

        hotModelPortItemsToGraphicalPortItem = {
            self._hotPipe.fromPort: self.fromPort,
            self._hotPipe.toPort: self.toPort
        }

        modelPortItemsToGraphicalPortItem = coldModelPortItemsToGraphicalPortItem | hotModelPortItemsToGraphicalPortItem
        return _mfs.InternalPiping([self._coldPipe, self._hotPipe], modelPortItemsToGraphicalPortItem)

    def exportPipeAndTeeTypesForTemp(self, startingUnit):  # pylint: disable=too-many-locals,too-many-statements
        unitNumber = startingUnit
        typeNr2 = 9511  # Temperature calculation from a pipe

        unitText = ""
        commentStars = 6 * "*"

        parameterNumber = 35
        inputNumbers = 6

        eqConst1 = 1
        eqConst3 = 3
        eqConst7 = 7
        eqConst8 = 8

        # Fixed strings
        initialValueS = "15.0 0.0 15.0 15.0 0.0 15.0"

        # Momentarily hardcoded
        equationNr = 6

        unitText += "UNIT " + str(unitNumber) + " TYPE " + str(typeNr2) + "\n"
        unitText += "!" + self.displayName + "\n"

        unitText += "PARAMETERS " + str(parameterNumber) + "\n"
        unitText += commentStars + " pipe and soil properties " + commentStars + "\n"
        unitText += self._addComment("dpLength", "! Length of buried pipe, m")
        unitText += self._addComment("dpDiamIn", "! Inner diameter of pipes, m")
        unitText += self._addComment("dpDiamOut", "! Outer diameter of pipes, m")
        unitText += self._addComment("dpLambda", "! Thermal conductivity of pipe material, kJ/(h*m*K)")
        unitText += self._addComment("dpDepth", "! Buried pipe depth, m")
        unitText += self._addComment("dpFlowMode", "! Direction of second pipe flow: 1 = same, 2 = opposite")
        unitText += self._addComment("dpDiamCase", "! Diameter of casing material, m")
        unitText += self._addComment("dpLambdaFill", "! Thermal conductivity of fill insulation, kJ/(h*m*K)")
        unitText += self._addComment("dpDistPtoP", "! Center-to-center pipe spacing, m")
        unitText += self._addComment("dpLambdaGap", "! Thermal conductivity of gap material, kJ/(h*m*K)")
        unitText += self._addComment("dpGapThick", "! Gap thickness, m")

        unitText += commentStars + " fluid properties " + commentStars + "\n"
        unitText += self._addComment("dpRhoFlu", "! Density of fluid, kg/m^3")
        unitText += self._addComment("dpLambdaFl", "! Thermal conductivity of fluid, kJ/(h*m*K)")
        unitText += self._addComment("dpCpFl", "! Specific heat of fluid, kJ/(kg*K)")
        unitText += self._addComment("dpViscFl", "! Viscosity of fluid, kg/(m*h)")

        unitText += commentStars + " initial conditions " + commentStars + "\n"
        unitText += self._addComment("dpTIniHot", "! Initial fluid temperature - Pipe hot, deg C")
        unitText += self._addComment("dpTIniCold", "! Initial fluid temperature - Pipe cold, deg C")

        unitText += commentStars + " thermal properties soil " + commentStars + "\n"
        unitText += self._addComment("dpLamdaSl", "! Thermal conductivity of soil, kJ/(h*m*K)")
        unitText += self._addComment("dpRhoSl", "! Density of soil, kg/m^3")
        unitText += self._addComment("dpCpSl", "! Specific heat of soil, kJ/(kg*K)")

        unitText += commentStars + " general temperature dependency (dependent on weather data) " + commentStars + "\n"
        unitText += self._addComment("TambAvg", "! Average surface temperature, deg C")
        unitText += self._addComment("dTambAmpl", "! Amplitude of surface temperature, deg C")
        unitText += self._addComment("ddTcwOffset", "! Days of minimum surface temperature")

        unitText += commentStars + " definition of nodes " + commentStars + "\n"
        unitText += self._addComment("dpNrFlNds", "! Number of fluid nodes")
        unitText += self._addComment("dpNrSlRad", "! Number of radial soil nodes")
        unitText += self._addComment("dpNrSlAx", "! Number of axial soil nodes")
        unitText += self._addComment("dpNrSlCirc", "! Number of circumferential soil nodes")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 1, m")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 2, m")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 3, m")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 4, m")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 5, m")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 6, m")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 7, m")
        unitText += self._addComment("dpRadNdDist", "! Radial distance of node 8, m")

        unitText += "INPUTS " + str(inputNumbers) + "\n"
        unitText += self._getInputs("Cold", self._coldPipe, self.toPort, self.fromPort)
        unitText += self._getInputs("Hot", self._hotPipe, self.fromPort, self.toPort)

        unitText += "***Initial values\n"
        unitText += initialValueS + "\n\n"

        unitText += "EQUATIONS " + str(equationNr) + "\n"
        unitText += self._getEquations("Cold", self._coldPipe, eqConst1, eqConst7, unitNumber)
        unitText += self._getEquations("Hot", self._hotPipe, eqConst3, eqConst8, unitNumber)

        unitNumber += 1
        unitText += "\n"

        return unitText, unitNumber

    def _getInputs(
            self,
            temperatureSuffix: str,
            pipe: _mfn.Pipe,
            pipeFromPort: _dppi.DoublePipePortItem,
            pipeToPort: _dppi.DoublePipePortItem
    ) -> str:
        mfrName = _helpers.getMfrName(pipe)

        fromBlockItem = pipeFromPort.parent
        firstColumn = f"T{fromBlockItem.displayName}{temperatureSuffix}"
        unitText = self._addComment(firstColumn, f"! Inlet fluid temperature - Pipe {temperatureSuffix}, deg C")

        firstColumn = f"{mfrName}"
        unitText += self._addComment(firstColumn, f"! Inlet fluid flow rate - Pipe {temperatureSuffix}, kg/h")

        toBlockItem = pipeToPort.parent
        firstColumn = f"T{toBlockItem.displayName}{temperatureSuffix}"
        unitText += self._addComment(firstColumn, "! Other side of pipe, deg C")

        return unitText

    def _getEquations(self, temperatureSuffix: str, pipe: _mfn.Pipe, equationConstant1, equationConstant2, unitNumber):
        firstColumn = (
            f"T{pipe.name} = [{unitNumber},{equationConstant1}]"
        )
        unitText = self._addComment(
            firstColumn, f"! {equationConstant1}: Outlet fluid temperature pipe {temperatureSuffix}, deg C"
        )

        mfrName = _helpers.getMfrName(pipe)
        firstColumn = f"Mfr{pipe.name} = {mfrName}"
        unitText += self._addComment(firstColumn, f"! Outlet mass flow rate pipe {temperatureSuffix}, kg/h")

        firstColumn = f"P{pipe.name}_kW = [{unitNumber},{equationConstant2}]/3600"
        unitText += self._addComment(
            firstColumn, f"! {equationConstant2}: Delivered energy pipe {temperatureSuffix}, kW"
        )

        return unitText

    @staticmethod
    def _addComment(firstColumn, comment):
        spacing = 40
        return str(firstColumn).ljust(spacing) + comment + "\n"

    def _updateModelPipes(self, displayName: str):
        coldFromPort: _mfn.PortItem = _dpmpi.ColdPortItem("Cold Input", _mfn.PortItemType.INPUT)
        coldToPort: _mfn.PortItem = _dpmpi.ColdPortItem("Cold Output", _mfn.PortItemType.OUTPUT)
        self._coldPipe = _mfn.Pipe(displayName + "Cold", self.childIds[0], coldFromPort, coldToPort)

        hotFromPort: _mfn.PortItem = _dpmpi.HotPortItem("Hot Input", _mfn.PortItemType.INPUT)
        hotToPort: _mfn.PortItem = _dpmpi.HotPortItem("Hot Output", _mfn.PortItemType.OUTPUT)
        self._hotPipe = _mfn.Pipe(displayName + "Hot", self.childIds[1], hotFromPort, hotToPort)


class DeleteDoublePipeConnectionCommand(_qtw.QUndoCommand):
    def __init__(self, doublePipeConnection: DoublePipeConnection, parentCommand: _qtw.QUndoCommand = None) -> None:
        super().__init__("Delete double pipe connection", parentCommand)
        self._connection = doublePipeConnection
        self._fromPort = self._connection.fromPort
        self._toPort = self._connection.toPort
        self._editor = self._connection.parent

    def undo(self):
        self._connection = DoublePipeConnection(self._fromPort, self._toPort, self._editor)

    def redo(self):
        self._connection.deleteConn()
        self._connection = None


@_dc.dataclass
class DoublePipeConnectionModel(_ser.UpgradableJsonSchemaMixinVersion0):  # pylint: disable=too-many-instance-attributes
    connectionId: int
    name: str
    id: int  # pylint: disable=invalid-name
    childIds: _tp.List[int]
    segmentsCorners: _tp.List[_tp.Tuple[float, float]]
    labelPos: _tp.Tuple[float, float]
    massFlowLabelPos: _tp.Tuple[float, float]
    fromPortId: int
    toPortId: int
    trnsysId: int

    @classmethod
    def from_dict(
        cls,
        data: _dcj.JsonDict,  # pylint: disable=duplicate-code  # 1
        validate=True,
        validate_enums: bool = True,
    ) -> "DoublePipeConnectionModel":
        data.pop(".__ConnectionDict__")
        doublePipeConnectionModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(DoublePipeConnectionModel, doublePipeConnectionModel)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,  # pylint: disable=duplicate-code # 1
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)
        data[".__ConnectionDict__"] = True
        return data

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("0810c9ea-85df-4431-bb40-3190c25c9161")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'trnsysGUI/hydraulicLoops/_dialogs/edit/_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.3
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_hydraulicLoopDialog(object):
    def setupUi(self, hydraulicLoopDialog):
        hydraulicLoopDialog.setObjectName("hydraulicLoopDialog")
        hydraulicLoopDialog.resize(554, 562)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(hydraulicLoopDialog.sizePolicy().hasHeightForWidth())
        hydraulicLoopDialog.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(hydraulicLoopDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.generalGroupBox = QtWidgets.QGroupBox(hydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.generalGroupBox.sizePolicy().hasHeightForWidth())
        self.generalGroupBox.setSizePolicy(sizePolicy)
        self.generalGroupBox.setObjectName("generalGroupBox")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.generalGroupBox)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.loopNameLabel = QtWidgets.QLabel(self.generalGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.loopNameLabel.sizePolicy().hasHeightForWidth())
        self.loopNameLabel.setSizePolicy(sizePolicy)
        self.loopNameLabel.setObjectName("loopNameLabel")
        self.horizontalLayout_2.addWidget(self.loopNameLabel)
        self.loopName = QtWidgets.QLineEdit(self.generalGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.loopName.sizePolicy().hasHeightForWidth())
        self.loopName.setSizePolicy(sizePolicy)
        self.loopName.setObjectName("loopName")
        self.horizontalLayout_2.addWidget(self.loopName)
        self.invalidNameWarningWidget = QtWidgets.QWidget(self.generalGroupBox)
        self.invalidNameWarningWidget.setMinimumSize(QtCore.QSize(22, 22))
        self.invalidNameWarningWidget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-warning.svg);")
        self.invalidNameWarningWidget.setObjectName("invalidNameWarningWidget")
        self.horizontalLayout_2.addWidget(self.invalidNameWarningWidget)
        spacerItem = QtWidgets.QSpacerItem(0, 19, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.fluidLabel = QtWidgets.QLabel(self.generalGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluidLabel.sizePolicy().hasHeightForWidth())
        self.fluidLabel.setSizePolicy(sizePolicy)
        self.fluidLabel.setObjectName("fluidLabel")
        self.horizontalLayout_2.addWidget(self.fluidLabel)
        self.fluidComboBox = QtWidgets.QComboBox(self.generalGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluidComboBox.sizePolicy().hasHeightForWidth())
        self.fluidComboBox.setSizePolicy(sizePolicy)
        self.fluidComboBox.setObjectName("fluidComboBox")
        self.horizontalLayout_2.addWidget(self.fluidComboBox)
        spacerItem1 = QtWidgets.QSpacerItem(130, 19, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_4.addLayout(self.horizontalLayout_2)
        self.useLoopWideDefaultsCheckBox = QtWidgets.QCheckBox(self.generalGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.useLoopWideDefaultsCheckBox.sizePolicy().hasHeightForWidth())
        self.useLoopWideDefaultsCheckBox.setSizePolicy(sizePolicy)
        self.useLoopWideDefaultsCheckBox.setObjectName("useLoopWideDefaultsCheckBox")
        self.verticalLayout_4.addWidget(self.useLoopWideDefaultsCheckBox)
        self.verticalLayout.addWidget(self.generalGroupBox)
        self.bulkOperationsGroupBox = QtWidgets.QGroupBox(hydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkOperationsGroupBox.sizePolicy().hasHeightForWidth())
        self.bulkOperationsGroupBox.setSizePolicy(sizePolicy)
        self.bulkOperationsGroupBox.setMinimumSize(QtCore.QSize(0, 180))
        self.bulkOperationsGroupBox.setStyleSheet("")
        self.bulkOperationsGroupBox.setObjectName("bulkOperationsGroupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.bulkOperationsGroupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.widget = QtWidgets.QWidget(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setMinimumSize(QtCore.QSize(20, 20))
        self.widget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-information.svg);")
        self.widget.setObjectName("widget")
        self.horizontalLayout_4.addWidget(self.widget)
        self.label_6 = QtWidgets.QLabel(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_6.sizePolicy().hasHeightForWidth())
        self.label_6.setSizePolicy(sizePolicy)
        self.label_6.setMinimumSize(QtCore.QSize(0, 20))
        self.label_6.setLineWidth(0)
        self.label_6.setWordWrap(True)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_4.addWidget(self.label_6)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.bulkLengthTypeComboBox = QtWidgets.QComboBox(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkLengthTypeComboBox.sizePolicy().hasHeightForWidth())
        self.bulkLengthTypeComboBox.setSizePolicy(sizePolicy)
        self.bulkLengthTypeComboBox.setObjectName("bulkLengthTypeComboBox")
        self.bulkLengthTypeComboBox.addItem("")
        self.bulkLengthTypeComboBox.addItem("")
        self.gridLayout.addWidget(self.bulkLengthTypeComboBox, 0, 0, 1, 1)
        self.bulkLengthLineEdit = QtWidgets.QLineEdit(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkLengthLineEdit.sizePolicy().hasHeightForWidth())
        self.bulkLengthLineEdit.setSizePolicy(sizePolicy)
        self.bulkLengthLineEdit.setObjectName("bulkLengthLineEdit")
        self.gridLayout.addWidget(self.bulkLengthLineEdit, 0, 1, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.bulkLengthInfoIconWidget = QtWidgets.QWidget(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkLengthInfoIconWidget.sizePolicy().hasHeightForWidth())
        self.bulkLengthInfoIconWidget.setSizePolicy(sizePolicy)
        self.bulkLengthInfoIconWidget.setMinimumSize(QtCore.QSize(20, 20))
        self.bulkLengthInfoIconWidget.setBaseSize(QtCore.QSize(0, 0))
        self.bulkLengthInfoIconWidget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-information.svg);")
        self.bulkLengthInfoIconWidget.setObjectName("bulkLengthInfoIconWidget")
        self.horizontalLayout.addWidget(self.bulkLengthInfoIconWidget)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 2, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 1, 0, 1, 1)
        self.bulkDiameterLineEdit = QtWidgets.QLineEdit(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkDiameterLineEdit.sizePolicy().hasHeightForWidth())
        self.bulkDiameterLineEdit.setSizePolicy(sizePolicy)
        self.bulkDiameterLineEdit.setObjectName("bulkDiameterLineEdit")
        self.gridLayout.addWidget(self.bulkDiameterLineEdit, 1, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_5.sizePolicy().hasHeightForWidth())
        self.label_5.setSizePolicy(sizePolicy)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 1, 2, 1, 1)
        self.bulkLengthUnitLabel = QtWidgets.QLabel(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkLengthUnitLabel.sizePolicy().hasHeightForWidth())
        self.bulkLengthUnitLabel.setSizePolicy(sizePolicy)
        self.bulkLengthUnitLabel.setObjectName("bulkLengthUnitLabel")
        self.gridLayout.addWidget(self.bulkLengthUnitLabel, 2, 0, 1, 1)
        self.bulkUValueLineEdit = QtWidgets.QLineEdit(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkUValueLineEdit.sizePolicy().hasHeightForWidth())
        self.bulkUValueLineEdit.setSizePolicy(sizePolicy)
        self.bulkUValueLineEdit.setObjectName("bulkUValueLineEdit")
        self.gridLayout.addWidget(self.bulkUValueLineEdit, 2, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 2, 2, 1, 1)
        self.horizontalLayout_5.addLayout(self.gridLayout)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem3)
        self.verticalLayout_2.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.bulkApplyButton = QtWidgets.QPushButton(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkApplyButton.sizePolicy().hasHeightForWidth())
        self.bulkApplyButton.setSizePolicy(sizePolicy)
        self.bulkApplyButton.setMaximumSize(QtCore.QSize(75, 24))
        self.bulkApplyButton.setObjectName("bulkApplyButton")
        self.horizontalLayout_3.addWidget(self.bulkApplyButton)
        self.bulkCancelButton = QtWidgets.QPushButton(self.bulkOperationsGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bulkCancelButton.sizePolicy().hasHeightForWidth())
        self.bulkCancelButton.setSizePolicy(sizePolicy)
        self.bulkCancelButton.setMaximumSize(QtCore.QSize(75, 24))
        self.bulkCancelButton.setObjectName("bulkCancelButton")
        self.horizontalLayout_3.addWidget(self.bulkCancelButton)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem4)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.verticalLayout.addWidget(self.bulkOperationsGroupBox)
        self.pipesGroupBox = QtWidgets.QGroupBox(hydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pipesGroupBox.sizePolicy().hasHeightForWidth())
        self.pipesGroupBox.setSizePolicy(sizePolicy)
        self.pipesGroupBox.setObjectName("pipesGroupBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.pipesGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.connectionsTableView = QtWidgets.QTableView(self.pipesGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.connectionsTableView.sizePolicy().hasHeightForWidth())
        self.connectionsTableView.setSizePolicy(sizePolicy)
        self.connectionsTableView.setObjectName("connectionsTableView")
        self.connectionsTableView.horizontalHeader().setStretchLastSection(True)
        self.verticalLayout_3.addWidget(self.connectionsTableView)
        self.pipesStatisticsLabel = QtWidgets.QLabel(self.pipesGroupBox)
        self.pipesStatisticsLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.pipesStatisticsLabel.setObjectName("pipesStatisticsLabel")
        self.verticalLayout_3.addWidget(self.pipesStatisticsLabel)
        self.verticalLayout.addWidget(self.pipesGroupBox)
        self.okCancelButtonBox = QtWidgets.QDialogButtonBox(hydraulicLoopDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.okCancelButtonBox.sizePolicy().hasHeightForWidth())
        self.okCancelButtonBox.setSizePolicy(sizePolicy)
        self.okCancelButtonBox.setOrientation(QtCore.Qt.Horizontal)
        self.okCancelButtonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.okCancelButtonBox.setObjectName("okCancelButtonBox")
        self.verticalLayout.addWidget(self.okCancelButtonBox)

        self.retranslateUi(hydraulicLoopDialog)
        self.okCancelButtonBox.accepted.connect(hydraulicLoopDialog.accept)
        self.okCancelButtonBox.rejected.connect(hydraulicLoopDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(hydraulicLoopDialog)

    def retranslateUi(self, hydraulicLoopDialog):
        _translate = QtCore.QCoreApplication.translate
        hydraulicLoopDialog.setWindowTitle(_translate("hydraulicLoopDialog", "Hydraulic loop properties"))
        self.generalGroupBox.setTitle(_translate("hydraulicLoopDialog", "Loop properties"))
        self.loopNameLabel.setText(_translate("hydraulicLoopDialog", "Name"))
        self.invalidNameWarningWidget.setToolTip(_translate("hydraulicLoopDialog", "<html><head/><body><p>The name is already taken</p></body></html>"))
        self.fluidLabel.setText(_translate("hydraulicLoopDialog", "Fluid"))
        self.useLoopWideDefaultsCheckBox.setText(_translate("hydraulicLoopDialog", "Use loop-wide defaults for pipe properties"))
        self.bulkOperationsGroupBox.setTitle(_translate("hydraulicLoopDialog", "Bulk pipe values assignment"))
        self.label_6.setText(_translate("hydraulicLoopDialog", "Values will be applied to all pipes. Unspecified properties will be left unchanged."))
        self.bulkLengthTypeComboBox.setItemText(0, _translate("hydraulicLoopDialog", "Total loop length"))
        self.bulkLengthTypeComboBox.setItemText(1, _translate("hydraulicLoopDialog", "Individual pipe length"))
        self.label.setText(_translate("hydraulicLoopDialog", "m"))
        self.bulkLengthInfoIconWidget.setToolTip(_translate("hydraulicLoopDialog", "<html><head/><body><p>The total length will be distributed evenly among the pipes</p></body></html>"))
        self.label_4.setText(_translate("hydraulicLoopDialog", "Diameter"))
        self.label_5.setText(_translate("hydraulicLoopDialog", "cm"))
        self.bulkLengthUnitLabel.setText(_translate("hydraulicLoopDialog", "U-Value"))
        self.label_3.setText(_translate("hydraulicLoopDialog", "W/(m^2 K)"))
        self.bulkApplyButton.setText(_translate("hydraulicLoopDialog", "Apply"))
        self.bulkCancelButton.setText(_translate("hydraulicLoopDialog", "Clear"))
        self.pipesGroupBox.setTitle(_translate("hydraulicLoopDialog", "Pipes"))
        self.pipesStatisticsLabel.setText(_translate("hydraulicLoopDialog", "Number of pipes: 5, total loop length: x"))



# pylint: skip-file
# type: ignore

import typing as tp
from math import sqrt

from PyQt5.QtGui import QColor, QPen
from PyQt5.QtWidgets import QGraphicsItemGroup, QGraphicsTextItem, QMenu

from trnsysGUI.CornerItem import CornerItem
from trnsysGUI.HorizSegmentMoveCommand import HorizSegmentMoveCommand

# This is needed to avoid a circular import but still be able to type check
if tp.TYPE_CHECKING:
    from trnsysGUI.connection.connectionBase import ConnectionBase


def calcDist(p1, p2):
    vec = p1 - p2
    norm = sqrt(vec.x() ** 2 + vec.y() ** 2)
    return norm


class SegmentItemBase(QGraphicsItemGroup):
    def __init__(self, startNode, endNode, parent: "ConnectionBase"):
        """
        A connection is displayed as a chain of segmentItems (stored in Connection.segments)
        Parameters.
        ----------
        startNode
        endNode
        parent: type(parent): Connection
        """

        super().__init__(None)
        self.logger = parent.logger

        self.setFlag(self.ItemIsSelectable, True)

        self.dragged = False
        self.initialised = False
        self.connection = parent

        self.firstChild = None
        self.secondChild = None
        self.cornerChild = None

        self.linePoints = None

        self.startNode = startNode
        self.endNode = endNode

        # These nodes are the nodes before and after the crossing
        self.start = None
        self.end = None

        # Unused. Related to interrupting segments for a clearer diagram
        self.disrBeforeNode = None
        self.disrAfterNode = None
        self.disrBeforeSeg = None
        self.disrAfterSeg = None
        self.disrBefore = False
        self.disrAfter = False
        self.hasBridge = False
        self.bridgedSegment = None

        # Only for editorMode 1
        self.firstLine = None
        self.secondLine = None
        self.secondCorner = None
        self.thirdCorner = None

        self.keyPr = 0
        # Used to only create the child objects once
        self._isDraggingInProgress = False

        self.insertInParentSegments()

        self.label = QGraphicsTextItem(self.connection.displayName)
        self.connection.parent.diagramScene.addItem(self.label)
        self.label.setVisible(False)
        self.label.setFlag(self.ItemIsMovable, True)
        self.labelMass = QGraphicsTextItem()
        self.connection.parent.diagramScene.addItem(self.labelMass)
        self.labelMass.setVisible(False)
        self.labelMass.setFlag(self.ItemIsMovable, True)

        self.setToolTip(self.connection.displayName)

    def segLength(self):
        return calcDist(self.line().p1(), self.line().p2())

    def interpolate(
        self,
        partLen2,
        totLenConn,
    ):
        # c1_r = 0
        # c1_b = 255
        c1_r = 160
        c1_b = 160
        c1_g = 160

        # c2_r = 255
        # c2_b = 0
        c2_r = 0
        c2_b = 0
        c2_g = 0

        try:
            f1 = int(partLen2 / totLenConn)
            f2 = int((totLenConn - partLen2) / totLenConn)
        except ZeroDivisionError:
            return QColor(100, 100, 100)
        else:
            return QColor(f1 * c2_r + f2 * c1_r, f1 * c2_g + f2 * c1_g, f1 * c2_b + f2 * c1_b)

    def line(self):
        return self.linePoints

    def setLine(self, *args):
        self.setZValue(-1)
        if len(args) == 2:
            p1, p2 = args
            x1 = p1.x()
            y1 = p1.y()
            x2 = p2.x()
            y2 = p2.y()
        else:
            x1, y1, x2, y2 = args

        self._setLineImpl(x1, y1, x2, y2)

    def _setLineImpl(self, x1, y1, x2, y2):
        raise NotImplementedError()

    def updateGrad(self):
        raise NotImplementedError()

    def insertInParentSegments(self):
        """
        This function inserts the segment in correct order to the segment list of the connection.
        Returns
        -------

        """
        prevSeg = None

        for s in self.connection.segments:
            if s.endNode is self.startNode:
                prevSeg = s

        # Todo: Add support for disr segments

        # if the startNode parent is a connection:
        if not hasattr(self.startNode.parent, "fromPort"):
            self.connection.segments.insert(self.connection.segments.index(prevSeg) + 1, self)
        else:

            self.connection.segments.insert(0, self)

    def mousePressEvent(self, e):

        if e.button() == 1:
            self.keyPr = 1
            self.logger.debug("Setting key to 1")

            self.connection.selectConnection()
            if self.isVertical():
                try:
                    self.oldX = self.startNode.parent.scenePos().x()
                except AttributeError:
                    pass
                else:
                    self.logger.debug("set oldx")

    def mouseMoveEvent(self, e):
        self.logger.debug(self.connection.parent.editorMode)
        if self.keyPr == 1:
            self.logger.debug("moved with button 1")
            newPos = e.pos()

            if self.connection.parent.editorMode == 0:
                if not self._isDraggingInProgress:
                    self.initInMode0()
                else:
                    self.dragInMode0(newPos)

            elif self.connection.parent.editorMode == 1:
                if type(self.startNode.parent) is CornerItem and type(self.endNode.parent) is CornerItem:
                    if not self.startNode.parent.isVisible():
                        self.startNode.parent.setVisible(True)
                    if not self.endNode.parent.isVisible():
                        self.endNode.parent.setVisible(True)
                    if self.isVertical():

                        self.logger.debug("Segment is vertical: %s", self.connection.segments.index(self))
                        self.endNode.parent.setPos(newPos.x(), self.endNode.parent.scenePos().y())
                        self.startNode.parent.setPos(newPos.x(), self.startNode.parent.scenePos().y())
                        self.updateGrad()

                    if self.isHorizontal():
                        self.logger.debug("Segment is horizontal")
                        self.endNode.parent.setPos(self.endNode.parent.scenePos().x(), newPos.y())
                        self.startNode.parent.setPos(self.startNode.parent.scenePos().x(), newPos.y())

                elif type(self.endNode.parent) is CornerItem and self.isVertical():
                    self.logger.debug("Segment is vertical and can't be moved")

                if self.isHorizontal():
                    isFirstSegment = hasattr(self.startNode.parent, "fromPort") and not self.startNode.prevN()
                    isLastSegment = hasattr(self.endNode.parent, "fromPort") and not self.endNode.nextN()

                    if isLastSegment:
                        self.logger.debug("A last segment is being dragged.")
                        if not self._isDraggingInProgress:
                            self._initInMode1(False)
                        self._dragInMode1(False, newPos)
                    elif isFirstSegment:
                        self.logger.debug("A first segment is being dragged.")
                        if not self._isDraggingInProgress:
                            self._initInMode1(True)
                        self._dragInMode1(True, newPos)
            else:
                self.logger.debug("Unrecognized editorMode in segmentItem mouseMoveEvent")

    def deleteNextHorizSeg(self, b, nextS):
        if b:
            pass
        else:
            nodeTodelete1 = self.endNode
            nodeTodelete2 = self.endNode.nextN()
            self.endNode = nextS.endNode

            self.startNode.setNext(self.endNode)
            self.endNode.setPrev(self.startNode)

            # x-position of the ending point of the next segment line

            posx1 = self.connection.segments[self.connection.segments.index(self) + 2].line().p2().x()

            self.connection.parent.diagramScene.removeItem(nextS)
            self.connection.segments.remove(nextS)
            self.connection.parent.diagramScene.removeItem(nodeTodelete1.parent)

            indexOfSelf = self.connection.segments.index(self)
            nextVS = self.connection.segments[indexOfSelf + 1]

            self.connection.parent.diagramScene.removeItem(nextVS)
            self.connection.segments.remove(nextVS)
            self.connection.parent.diagramScene.removeItem(nodeTodelete2.parent)

            self.setLine(
                self.startNode.parent.scenePos().x(),
                self.startNode.parent.scenePos().y(),
                posx1,
                self.startNode.parent.scenePos().y(),
            )

    def deletePrevHorizSeg(self, b, prevS):
        if b:
            pass
        else:
            nodeTodelete1 = self.startNode
            nodeTodelete2 = self.startNode.prevN()
            self.startNode = prevS.startNode

            self.startNode.setNext(self.endNode)
            self.endNode.setPrev(self.startNode)

            posx1 = self.connection.segments[self.connection.segments.index(self) - 2].line().p1().x()

            self.connection.parent.diagramScene.removeItem(prevS)
            self.connection.segments.remove(prevS)
            self.connection.parent.diagramScene.removeItem(nodeTodelete1.parent)

            indexOfSelf = self.connection.segments.index(self)
            nextVS = self.connection.segments[indexOfSelf - 1]

            self.connection.parent.diagramScene.removeItem(nextVS)
            self.connection.segments.remove(nextVS)
            self.connection.parent.diagramScene.removeItem(nodeTodelete2.parent)

            self.setLine(
                posx1,
                self.endNode.parent.scenePos().y(),
                self.endNode.parent.scenePos().x(),
                self.endNode.parent.scenePos().y(),
            )

    def deleteSegment(self):
        nodeToConnect = self.startNode.prevN()
        nodeToConnect2 = self.endNode.nextN()

        nodeToConnect.setNext(nodeToConnect2)

        self.connection.parent.diagramScene.removeItem(self)
        self.connection.segments.remove(self)
        self.connection.parent.diagramScene.removeItem(self.startNode.parent)
        self.connection.parent.diagramScene.removeItem(self.endNode.parent)

    def splitSegment(self):
        pass

    def mouseReleaseEvent(self, e):
        # Should be same as below
        # self.scene().removeItem(self)

        if e.button() == 1:
            self.keyPr = 0

            if self.connection.parent.editorMode == 0:
                if self._isDraggingInProgress:
                    self.cornerChild.setFlag(self.ItemSendsScenePositionChanges, True)

                    self.hide()
                    self.connection.segments.remove(self)
                    self.connection.parent.diagramScene.removeItem(self)

            elif self.connection.parent.editorMode == 1:
                if self.isVertical():
                    try:
                        self.oldX
                    except AttributeError:
                        pass
                    else:
                        command = HorizSegmentMoveCommand(self, self.oldX, "Moving segment command")

                        self.connection.parent.parent().undoStack.push(command)
                        self.oldX = self.scenePos().x()

                if self.isHorizontal():
                    if type(self.startNode.parent) is CornerItem and type(self.endNode.parent) is CornerItem:
                        try:

                            nextHorizSeg = self.connection.segments[self.connection.segments.index(self) + 2]
                            prevHorizSeg = self.connection.segments[self.connection.segments.index(self) - 2]
                        except IndexError:
                            self.logger.debug("no next or prev segments")
                        else:
                            if nextHorizSeg.isHorizontal() and int(self.endNode.parent.pos().y() - 10) <= int(
                                nextHorizSeg.line().p2().y()
                            ) <= int(self.endNode.parent.pos().y() + 10):
                                self.deleteNextHorizSeg(False, nextHorizSeg)
                                self.logger.debug("next horizontal")
                                return

                            if prevHorizSeg.isHorizontal() and int(self.startNode.parent.pos().y() - 10) <= int(
                                prevHorizSeg.line().p2().y()
                            ) <= int(self.startNode.parent.pos().y() + 10):
                                self.deletePrevHorizSeg(False, prevHorizSeg)
                                self.logger.debug("previous horizontal")
                                return

                if self.secondCorner is not None:
                    self.logger.debug("Second corner is not none")
                    # if PortItem
                    if hasattr(self.endNode.parent, "fromPort"):
                        segbef = self.connection.segments[self.connection.getNodePos(self.secondCorner.node.prevN().parent)]

                        segbef.setLine(
                            segbef.line().p1().x(),
                            segbef.line().p1().y(),
                            segbef.line().p2().x(),
                            self.secondCorner.scenePos().y(),
                        )
                        self.setLine(
                            self.thirdCorner.scenePos().x(),
                            self.thirdCorner.scenePos().y(),
                            self.line().p2().x(),
                            self.line().p2().y(),
                        )
                        self.secondCorner.setFlag(self.ItemSendsScenePositionChanges, True)
                        self.thirdCorner.setFlag(self.ItemSendsScenePositionChanges, True)

                        # Allow for iterative branching
                        self.secondCorner = None
                        self.thirdCorner = None
                        self.firstLine = None
                        self.secondLine = None
                        self._isDraggingInProgress = False

                    # if PortItem
                    elif hasattr(self.startNode.parent, "fromPort"):

                        segafter = self.connection.segments[self.connection.getNodePos(self.thirdCorner.node.nextN().parent)]

                        segafter.setLine(
                            segafter.line().p1().x(),
                            self.thirdCorner.scenePos().y(),
                            segafter.line().p2().x(),
                            segafter.line().p2().y(),
                        )
                        self.setLine(
                            self.line().p1().x(),
                            self.line().p1().y(),
                            self.secondCorner.scenePos().x(),
                            self.secondCorner.scenePos().y(),
                        )

                        self.secondCorner.setFlag(self.ItemSendsScenePositionChanges, True)
                        self.thirdCorner.setFlag(self.ItemSendsScenePositionChanges, True)

                        # Allow for iterative branching
                        self.secondCorner = None
                        self.thirdCorner = None
                        self.firstLine = None
                        self.secondLine = None
                        self._isDraggingInProgress = False

                    else:
                        self.logger.debug("getting no start or end")

                else:
                    self.logger.debug("Second corner is none")
            else:
                pass

    def initInMode0(self):
        if (hasattr(self.startNode.parent, "fromPort")) and (self.startNode.prevN() is not None):
            self.disrAfterNode = self.startNode
            self.start = self.startNode.prevN().prevN()


            segments = self.connection.segments
            for s in segments:
                if s.startNode is self.start:
                    self.disrBeforeSeg = s

            self.disrAfterSeg = self
            self.disrBefore = True

        else:
            self.start = self.startNode

        if (hasattr(self.endNode.parent, "fromPort")) and (self.endNode.nextN() is not None):
            self.disrBeforeNode = self.endNode
            self.end = self.endNode.nextN().nextN()


            segments = self.connection.segments
            for s in segments:
                if s.endNode is self.end:
                    self.disrAfterSeg = s

            self.disrBeforeSeg = self
            self.disrAfter = True

        else:
            self.end = self.endNode

        rad = self.connection.getRadius()

        self.cornerChild = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.start, self.end, self.connection)
        self.firstChild = self._createSegment(self.start, self.cornerChild.node)
        self.secondChild = self._createSegment(self.cornerChild.node, self.end)

        self.start.setNext(self.cornerChild.node)
        self.end.setPrev(self.cornerChild.node)

        self.firstChild.setVisible(False)
        self.secondChild.setVisible(False)
        self.cornerChild.setVisible(False)

        self.connection.parent.diagramScene.addItem(self.firstChild)
        self.connection.parent.diagramScene.addItem(self.secondChild)
        self.connection.parent.diagramScene.addItem(self.cornerChild)

        self._isDraggingInProgress = True

    def _initInMode1(self, b):

        rad = self.connection.getRadius()

        if b:
            if (hasattr(self.startNode.parent, "fromPort")) and (self.startNode.prevN() is None):
                # We are at the toPort.
                # self.end = self.endNode
                # self.start = self.startNode

                self.secondCorner = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self.connection)
                self.thirdCorner = CornerItem(

                    -rad, -rad, 2 * rad, 2 * rad, self.secondCorner.node, self.endNode, self.connection
                )

                self.secondCorner.node.setNext(self.thirdCorner.node)
                self.startNode.setNext(self.secondCorner.node)
                self.endNode.setPrev(self.thirdCorner.node)

                self.endNode = self.secondCorner.node

                self.firstLine = self._createSegment(self.secondCorner.node, self.thirdCorner.node)
                self.secondLine = self._createSegment(self.thirdCorner.node, self.thirdCorner.node.nextN())

                self.secondCorner.setVisible(False)
                self.thirdCorner.setVisible(False)
                self.firstLine.setVisible(False)
                self.secondLine.setVisible(False)
                # self.thirdLine.setVisible(False)

                self.connection.parent.diagramScene.addItem(self.secondCorner)
                self.connection.parent.diagramScene.addItem(self.thirdCorner)
                self.connection.parent.diagramScene.addItem(self.firstLine)
                self.connection.parent.diagramScene.addItem(self.secondLine)
                self.logger.debug("inited")

                self._isDraggingInProgress = True
        else:
            if (hasattr(self.endNode.parent, "fromPort")) and (self.endNode.nextN() is None):
                # We are at the toPort.
                # self.end = self.endNode
                # self.start = self.startNode

                self.secondCorner = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self.connection)
                self.thirdCorner = CornerItem(

                    -rad, -rad, 2 * rad, 2 * rad, self.secondCorner.node, self.endNode, self.connection
                )

                self.secondCorner.node.setNext(self.thirdCorner.node)
                self.startNode.setNext(self.secondCorner.node)
                self.endNode.setPrev(self.thirdCorner.node)

                self.startNode = self.thirdCorner.node

                self.firstLine = self._createSegment(self.secondCorner.node.prevN(), self.secondCorner.node)
                self.secondLine = self._createSegment(self.secondCorner.node, self.thirdCorner.node)

                self.secondCorner.setVisible(False)
                self.thirdCorner.setVisible(False)
                self.firstLine.setVisible(False)
                self.secondLine.setVisible(False)
                # self.thirdLine.setVisible(False)

                self.connection.parent.diagramScene.addItem(self.secondCorner)
                self.connection.parent.diagramScene.addItem(self.thirdCorner)
                self.connection.parent.diagramScene.addItem(self.firstLine)
                self.connection.parent.diagramScene.addItem(self.secondLine)
                self.logger.debug("inited")

                self._isDraggingInProgress = True

    def _createSegment(self, startNode, endNode) -> "SegmentItemBase":
        raise NotImplementedError()


    def isVertical(self):
        return self.line().p1().x() == self.line().p2().x()

    def isHorizontal(self):
        return self.line().p1().y() == self.line().p2().y()

    def dragInMode0(self, newPos):
        p1 = self.line().p1()
        p2 = self.line().p2()

        if len(self.scene().items(newPos)) == 0:

            self.firstChild.setLine(p1.x(), p1.y(), newPos.x(), newPos.y())
            self.secondChild.setLine(newPos.x(), newPos.y(), p2.x(), p2.y())

            self.cornerChild.setPos(newPos)

            self.firstChild.updateGrad()
            self.secondChild.updateGrad()

            # Bring corner to front
            self.cornerChild.setZValue(100)
            self.firstChild.setZValue(1)
            self.secondChild.setZValue(1)

            self.firstChild.setVisible(True)
            self.secondChild.setVisible(True)
            self.cornerChild.setVisible(True)

    def _dragInMode1(self, b, newPos):
        self.logger.debug("after inited")

        if b:
            self.thirdCorner.setPos(newPos.x() - 10, newPos.y())

            self.secondCorner.setPos(newPos.x() - 10, self.connection.fromPort.scenePos().y())
            self.thirdCorner.node.nextN().parent.setY(newPos.y())

            self.firstLine.setLine(
                self.secondCorner.scenePos().x(),
                self.secondCorner.scenePos().y(),
                self.thirdCorner.scenePos().x(),
                newPos.y(),
            )
            self.secondLine.setLine(
                self.thirdCorner.scenePos().x(),
                self.thirdCorner.scenePos().y(),
                self.thirdCorner.node.nextN().parent.scenePos().x(),
                self.thirdCorner.node.nextN().parent.scenePos().y(),
            )
            self.setLine(
                self.startNode.parent.fromPort.scenePos().x(),
                self.startNode.parent.fromPort.scenePos().y(),
                self.secondCorner.scenePos().x(),
                self.secondCorner.scenePos().y(),
            )

            self.secondCorner.setZValue(100)
            self.thirdCorner.setZValue(100)
            self.firstLine.setZValue(1)
            self.secondLine.setZValue(1)

            self.secondCorner.setVisible(True)
            self.thirdCorner.setVisible(True)
            self.firstLine.setVisible(True)
            self.secondLine.setVisible(True)

        else:
            self.secondCorner.setPos(newPos.x() + 10, newPos.y())

            self.thirdCorner.setPos(newPos.x() + 10, self.connection.toPort.scenePos().y())
            self.secondCorner.node.prevN().parent.setY(newPos.y())

            self.firstLine.setLine(
                self.secondCorner.node.prevN().parent.scenePos().x(),
                newPos.y(),
                self.secondCorner.scenePos().x(),
                newPos.y(),
            )
            self.secondLine.setLine(
                self.secondCorner.scenePos().x(),
                self.secondCorner.scenePos().y(),
                self.thirdCorner.scenePos().x(),
                self.thirdCorner.scenePos().y(),
            )
            self.setLine(
                self.thirdCorner.scenePos().x(),
                self.thirdCorner.scenePos().y(),
                self.endNode.parent.toPort.scenePos().x(),
                self.endNode.parent.toPort.scenePos().y(),
            )

            self.secondCorner.setZValue(100)
            self.thirdCorner.setZValue(100)
            self.firstLine.setZValue(1)
            self.secondLine.setZValue(1)

            self.secondCorner.setVisible(True)
            self.thirdCorner.setVisible(True)
            self.firstLine.setVisible(True)
            self.secondLine.setVisible(True)

    def renameConn(self):
        self.scene().parent().showSegmentDlg(self)

    def printItemsAt(self):
        self.logger.debug("Items at startnode are %s", str(self.scene().items(self.line().p1())))
        self.logger.debug("Items at endnode are %s", str(self.scene().items(self.line().p2())))

        for s in self.connection.segments:
            self.logger.debug(
                "Segment in list is %s has startnode %s endnode %s",
                str(s),
                str(s.startNode.parent),
                str(s.endNode.parent),
            )

    def contextMenuEvent(self, event):
        menu = self._getContextMenu()

        menu.exec(event.screenPos())

    def _getContextMenu(self) -> QMenu:
        menu = QMenu()
        a1 = menu.addAction("Rename...")
        a1.triggered.connect(self.renameConn)
        a2 = menu.addAction("Delete this connection")
        a2.triggered.connect(self.connection.createDeleteUndoCommandAndAddToStack)
        a3 = menu.addAction("Invert this connection")
        a3.triggered.connect(self.connection.invertConnection)
        a4 = menu.addAction("Toggle name")
        a4.triggered.connect(self.connection.toggleLabelVisible)
        a5 = menu.addAction("Toggle mass flow")
        a5.triggered.connect(self.connection.toggleMassFlowLabelVisible)
        return menu

    def setLabelVisible(self, isVisible: bool) -> None:
        self.label.setVisible(isVisible)

    def toggleLabelVisible(self) -> None:
        wasVisible = self.label.isVisible()
        self.setLabelVisible(not wasVisible)

    def setMassFlowLabelVisible(self, isVisible: bool) -> None:
        self.labelMass.setVisible(isVisible)

    def toggleMassFlowLabelVisible(self) -> None:
        wasVisible = self.labelMass.isVisible()
        self.setMassFlowLabelVisible(not wasVisible)

    def setSelect(self, isSelected: bool) -> None:
        raise NotImplementedError()

    @staticmethod
    def _createSelectPen() -> QPen:
        color = QColor(125, 242, 189)
        width = 4

        selectPen = QPen(color, width)
        return selectPen

    def setColorAndWidthAccordingToMassflow(self, color, width):
        raise NotImplementedError()



import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import dataclasses_jsonschema as _dcj

import pytrnsys.utils.serialization as _ser
import trnsysGUI.BlockItem as _bi
import trnsysGUI.doublePipeModelPortItems as _dpmpi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn


class DoublePipeConnectorBase(_bi.BlockItem, _mfs.MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.w = 40
        self.h = 20

        self.childIds = []
        self.childIds.append(self.trnsysId)
        self.childIds.append(self.parent.parent().idGen.getTrnsysID())

        self._updateModelPipes(self.displayName)

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        raise NotImplementedError()

    def rotateBlockCW(self):
        super().rotateBlockCW()
        self._flipPipes()

    def rotateBlockCCW(self):
        super().rotateBlockCCW()
        self._flipPipes()

    def resetRotation(self):
        super().resetRotation()
        self.updateFlipStateV(0)

    def encode(self):
        portListInputs = []
        portListOutputs = []

        for inp in self.inputs:
            portListInputs.append(inp.id)
        for output in self.outputs:
            portListOutputs.append(output.id)

        blockPosition = (float(self.pos().x()), float(self.pos().y()))

        connectorModel = DoublePipeBlockItemModel(
            self.name,
            self.displayName,
            blockPosition,
            self.id,
            self.trnsysId,
            self.childIds,
            portListInputs,
            portListOutputs,
            self.flippedH,
            self.flippedV,
            self.rotationN,
        )

        dictName = "Block-"
        return dictName, connectorModel.to_dict()

    def decode(self, i, resBlockList):
        model = DoublePipeBlockItemModel.from_dict(i)

        self.setDisplayName(model.BlockDisplayName)
        self.setPos(float(model.blockPosition[0]), float(model.blockPosition[1]))
        self.id = model.id
        self.trnsysId = model.trnsysId
        self.childIds = model.childIds

        for index, inp in enumerate(self.inputs):
            inp.id = model.portsIdsIn[index]

        for index, out in enumerate(self.outputs):
            out.id = model.portsIdsOut[index]

        self.updateFlipStateH(model.flippedH)
        self.updateFlipStateV(model.flippedV)
        self.rotateBlockToN(model.rotationN)

        resBlockList.append(self)

    def getInternalPiping(self) -> _mfs.InternalPiping:
        raise NotImplementedError()

    def _updateModelPipes(self, displayName: str) -> None:
        coldInput = _dpmpi.ColdPortItem("coldInput", _mfn.PortItemType.INPUT)
        coldOutput = _dpmpi.ColdPortItem("coldOutput", _mfn.PortItemType.OUTPUT)
        self._coldPipe = _mfn.Pipe(displayName + "Cold", self.childIds[0], coldInput, coldOutput)

        hotInput = _dpmpi.HotPortItem("hotInput", _mfn.PortItemType.INPUT)
        hotOutput = _dpmpi.HotPortItem("hotOutput", _mfn.PortItemType.OUTPUT)
        self._hotPipe = _mfn.Pipe(displayName + "Hot", self.childIds[1], hotInput, hotOutput)

    def setDisplayName(self, newName):
        super().setDisplayName(newName)
        self._updateModelPipes(newName)

    def _flipPipes(self):
        angle = (self.rotationN % 4) * 90
        if angle == 0:
            self.updateFlipStateV(False)
        elif angle == 90:
            self.updateFlipStateV(True)
        elif angle == 180:
            self.updateFlipStateV(True)
        elif angle == 270:
            self.updateFlipStateV(False)


@_dc.dataclass
class DoublePipeBlockItemModel(_ser.UpgradableJsonSchemaMixinVersion0):  # pylint: disable=too-many-instance-attributes
    BlockName: str  # pylint: disable=invalid-name
    BlockDisplayName: str  # pylint: disable=invalid-name
    blockPosition: _tp.Tuple[float, float]
    id: int  # pylint: disable=invalid-name
    trnsysId: int
    childIds: _tp.List[int]
    portsIdsIn: _tp.List[int]
    portsIdsOut: _tp.List[int]
    flippedH: bool
    flippedV: bool
    rotationN: int

    @classmethod
    def from_dict(
        cls,
        data: _dcj.JsonDict,
        validate=True,  # pylint: disable=duplicate-code
        validate_enums: bool = True,
    ) -> "DoublePipeBlockItemModel":
        data.pop(".__BlockDict__")
        doublePipeBlockItemModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(DoublePipeBlockItemModel, doublePipeBlockItemModel)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)  # pylint: disable=duplicate-code
        data[".__BlockDict__"] = True
        return data

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("e5149c30-9f05-4a3a-8a3c-9ada74143802")



import dataclasses as _dc
import os as _os
import random as _rnd
import shutil as _sh
import typing as _tp

from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QMenu, QMessageBox, QTreeView

import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
import trnsysGUI.storageTank.model as _model
import trnsysGUI.storageTank.side as _sd
from trnsysGUI import idGenerator as _id
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.HeatExchanger import HeatExchanger  # type: ignore[attr-defined]
from trnsysGUI.MyQFileSystemModel import MyQFileSystemModel  # type: ignore[attr-defined]
from trnsysGUI.MyQTreeView import MyQTreeView  # type: ignore[attr-defined]
from trnsysGUI.directPortPair import DirectPortPair
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin
from trnsysGUI.singlePipePortItem import SinglePipePortItem
from trnsysGUI.storageTank.ConfigureStorageDialog import ConfigureStorageDialog
from trnsysGUI.type1924.createType1924 import Type1924_TesPlugFlow  # type: ignore[attr-defined]

InOut = _tp.Literal["In", "Out"]
_T = _tp.TypeVar("_T", covariant=True)


@_dc.dataclass
class PortIds:
    inputId: int
    outputId: int


class StorageTank(BlockItem,
                  MassFlowNetworkContributorMixin):
    # pylint: disable=too-many-instance-attributes,too-many-public-methods
    HEAT_EXCHANGER_WIDTH = 40

    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.parent = parent
        self._idGenerator: _id.IdGenerator = self.parent.parent().idGen
        self.dckFilePath = ""

        self.directPortPairs: _tp.List[DirectPortPair] = []

        self.heatExchangers: _tp.List[HeatExchanger] = []

        self.blackBoxEquations = []

        self.nTes = self.parent.parent().idGen.getStoragenTes()
        self.storageType = self.parent.parent().idGen.getStorageType()

        self.changeSize()

        self.path = None
        self.addTree()

    @property
    def leftDirectPortPairsPortItems(self):
        return self._getDirectPortPairPortItems(_sd.Side.LEFT)

    @property
    def rightDirectPortPairsPortItems(self):
        return self._getDirectPortPairPortItems(_sd.Side.RIGHT)

    def _getDirectPortPairPortItems(self, side: _sd.Side):
        return [
            p for dpp in self.directPortPairs if dpp.side == side for p in [dpp.fromPort, dpp.toPort]
        ]

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.STORAGE_TANK_SVG

    # Setter functions
    def setParent(self, p):
        self.logger.debug("Setting parent of Storage Tank (and its hx)")
        self.parent = p

        if self not in self.parent.parent().trnsysObj:
            self.parent.parent().trnsysObj.append(self)

        for heatExchanger in self.heatExchangers:
            heatExchanger.parent = self

    def addDirectPortPair(  # pylint: disable=too-many-arguments
            self,
            trnsysId: int,
            side: _sd.Side,
            relativeInputHeight: float,
            relativeOutputHeight: float,
            storageTankHeight: float,
            portIds: _tp.Optional[PortIds] = None,
    ):
        inputPort = self._createPort("i", relativeInputHeight, storageTankHeight, side)
        outputPort = self._createPort("o", relativeOutputHeight, storageTankHeight, side)

        randomInt = int(_rnd.uniform(20, 200))
        randomColor = QColor(randomInt, randomInt, randomInt)
        self._setPortColor(inputPort, randomColor)
        self._setPortColor(outputPort, randomColor)

        if portIds:
            inputPort.id = portIds.inputId
            outputPort.id = portIds.outputId

        directPortPair = DirectPortPair(
            trnsysId, inputPort, outputPort, relativeInputHeight, relativeOutputHeight, side
        )

        self.directPortPairs.append(directPortPair)
        self.inputs.append(directPortPair.fromPort)
        self.outputs.append(directPortPair.toPort)

    def _createPort(
            self, name: str, relativeHeight: float, storageTankHeight: float, side: _sd.Side
    ) -> SinglePipePortItem:
        sideNr = side.toSideNr()
        portItem = _cspi.createSinglePipePortItem(name, sideNr, self)
        portItem.setZValue(100)
        xPos = 0 if side == _sd.Side.LEFT else self.w
        yPos = storageTankHeight - relativeHeight * storageTankHeight
        portItem.setPos(xPos, yPos)
        portItem.side = sideNr
        return portItem

    @staticmethod
    def _setPortColor(portItem: SinglePipePortItem, color: QColor) -> None:
        portItem.innerCircle.setBrush(color)
        portItem.visibleColor = color

    def addHeatExchanger(self, name, trnsysId, side, relativeInputHeight, relativeOutputHeight):
        heatExchanger = HeatExchanger(
            trnsysId=trnsysId,
            sideNr=side.toSideNr(),
            width=self.HEAT_EXCHANGER_WIDTH,
            relativeInputHeight=relativeInputHeight,
            relativeOutputHeight=relativeOutputHeight,
            storageTankWidth=self.w,
            storageTankHeight=self.h,
            parent=self,
            name=name,
        )
        return heatExchanger

    def updateImage(self):
        super().updateImage()
        self.label.setPos(self.label.pos().x(), self.h)

    def updatePortItemPositions(self, deltaH, deltaW):
        for portItem in self.inputs + self.outputs:
            oldRelativeHeight = portItem.pos().y() / self.h
            if portItem.side == 0:
                portItem.setPos(portItem.pos().x(), oldRelativeHeight * (self.h + deltaH))
            else:
                portItem.setPos(portItem.pos().x() + deltaW, oldRelativeHeight * (self.h + deltaH))

    def updateHeatExchangersAfterTankSizeChange(self):
        for heatExchanger in self.heatExchangers:
            heatExchanger.setTankSize(self.w, self.h)

    def encode(self):
        if not self.isVisible():
            raise RuntimeError("Cannot encode an invisible storage tank.")

        heatExchangerModels = self._getHeatExchangerModelsForEncode()
        portPairModels = self._getDirectPortPairModelsForEncode()
        position = float(self.pos().x()), float(self.pos().y())

        storageTankModel = _model.StorageTank(
            self.flippedH,
            self.flippedV,
            self.name,
            self.displayName,
            self.id,
            self.trnsysId,
            self.h,
            position,
            heatExchangerModels,
            portPairModels,
        )

        dictName = "Block-"
        return dictName, storageTankModel.to_dict()

    def _getDirectPortPairModelsForEncode(self):
        portPairModels = []
        for directPort in self.directPortPairs:
            side = _sd.Side.createFromSideNr(directPort.fromPort.side)

            inputPortModel = _model.Port(directPort.fromPort.id, directPort.relativeInputHeight)

            outputPortModel = _model.Port(directPort.toPort.id, directPort.relativeOutputHeight)

            portPairModel = _model.PortPair(side, directPort.trnsysId, inputPortModel, outputPortModel)

            directPortPairModel = _model.DirectPortPair(portPairModel)

            portPairModels.append(directPortPairModel)

        return portPairModels

    def _getHeatExchangerModelsForEncode(self):
        heatExchangerModels = []
        for heatExchanger in self.heatExchangers:
            side = _sd.Side.createFromSideNr(heatExchanger.sSide)

            inputPort = _model.Port(
                heatExchanger.port1.id,
                heatExchanger.relativeInputHeight,
            )

            outputPort = _model.Port(
                heatExchanger.port2.id,
                heatExchanger.relativeOutputHeight,
            )

            portPair = _model.PortPair(side, heatExchanger.trnsysId, inputPort, outputPort)

            heatExchangerModel = _model.HeatExchanger(
                portPair, heatExchanger.displayName, heatExchanger.w, self.id, heatExchanger.id
            )

            heatExchangerModels.append(heatExchangerModel)

        return heatExchangerModels

    def decode(self, i, resBlockList):
        offsetX = 0
        offsetY = 0
        self._decodeInternal(i, offsetX, offsetY, resBlockList, shallSetNamesAndIDs=True)

    def _decodeInternal(  # pylint: disable=too-many-arguments
            self,
            i,
            offsetX,
            offsetY,
            resBlockList,
            shallSetNamesAndIDs: bool,
    ):
        self.logger.debug("Loading a Storage in Decoder")

        model = _model.StorageTank.from_dict(i)

        self.flippedH = model.isHorizontallyFlipped

        if shallSetNamesAndIDs:
            self.displayName = model.BlockDisplayName

        self.changeSize()
        self.h = model.height
        self.updateImage()

        self.setPos(model.position[0] + offsetX, model.position[1] + offsetY)

        if shallSetNamesAndIDs:
            self.trnsysId = model.trnsysId
            self.id = model.id

        for heatExchangerModel in model.heatExchangers:
            self._decodeHeatExchanger(heatExchangerModel, shallSetNamesAndIDs)

        for portPairModel in model.directPortPairs:
            self._decodeDirectPortPair(portPairModel)

        resBlockList.append(self)

    def _decodeDirectPortPair(
            self,
            portPairModel: _model.DirectPortPair,
    ) -> None:
        portPair = portPairModel.portPair

        portIds = PortIds(portPair.inputPort.id, portPair.outputPort.id)

        self.addDirectPortPair(
            portPair.trnsysId,
            portPair.side,
            portPair.inputPort.relativeHeight,
            portPair.outputPort.relativeHeight,
            storageTankHeight=self.h,
            portIds=portIds,
        )

    def _decodeHeatExchanger(self, heatExchangerModel: _model.HeatExchanger, shallSetNamesAndIDs: bool):
        portPair = heatExchangerModel.portPair

        nameSuffix = "" if shallSetNamesAndIDs else "New"
        name = heatExchangerModel.name + nameSuffix

        heatExchanger = self.addHeatExchanger(
            name, portPair.trnsysId, portPair.side, portPair.inputPort.relativeHeight,
            portPair.outputPort.relativeHeight
        )

        if shallSetNamesAndIDs:
            heatExchanger.setId(heatExchangerModel.id)

        heatExchanger.port1.id = portPair.inputPort.id
        heatExchanger.port2.id = portPair.outputPort.id

    def decodePaste(  # pylint: disable=too-many-arguments
            self, i, offset_x, offset_y, resConnList, resBlockList, **kwargs
    ):
        self._decodeInternal(i, offset_x, offset_y, resBlockList, shallSetNamesAndIDs=False)

    def getTemperatureVariableName(self, portItem: SinglePipePortItem) -> str:
        directPortPair = self._getDirectPortPairForPortItemOrNone(portItem)
        if directPortPair:
            return self._getTemperatureVariableNameForDirectPortPairPortItem(directPortPair, portItem)

        heatExchanger = self._getHeatExchangerForPortItem(portItem)
        if heatExchanger:
            return self._getTemperatureVariableNameForHeatExchangerPortItem(heatExchanger)

        raise ValueError("Port item doesn't belong to this storage tank.")

    def getFlowSolverParametersId(self, portItem: SinglePipePortItem) -> int:
        directPortPair = self._getDirectPortPairForPortItemOrNone(portItem)
        if directPortPair:
            return directPortPair.trnsysId

        heatExchanger = self._getHeatExchangerForPortItem(portItem)
        if heatExchanger:
            return heatExchanger.trnsysId

        raise ValueError("Port item doesn't belong to this storage tank.")

    def assignIDsToUninitializedValuesAfterJsonFormatMigration(
            self, generator: _id.IdGenerator
    ) -> None:  # type: ignore[attr-defined]
        for heatExchanger in self.heatExchangers:
            if heatExchanger.trnsysId == generator.UNINITIALIZED_ID:
                heatExchanger.trnsysId = generator.getTrnsysID()

        for directPortPair in self.directPortPairs:
            if directPortPair.trnsysId == generator.UNINITIALIZED_ID:
                directPortPair.trnsysId = generator.getTrnsysID()

    def _getHeatExchangerForPortItem(self, portItem: SinglePipePortItem) -> _tp.Optional[HeatExchanger]:
        heatExchanger = self._getSingleOrNone(hx for hx in self.heatExchangers if portItem in [hx.port1, hx.port2])

        return heatExchanger

    def _getDirectPortPairForPortItemOrNone(self, portItem: SinglePipePortItem) -> _tp.Optional[DirectPortPair]:
        directPortPair = self._getSingleOrNone(
            dpp for dpp in self.directPortPairs if portItem in [dpp.fromPort, dpp.toPort]
        )

        return directPortPair

    @staticmethod
    def _getSingleOrNone(iterable: _tp.Iterable[_T]) -> _tp.Optional[_T]:
        sequence = list(iterable)

        if not sequence:
            return None

        if len(sequence) > 1:
            raise ValueError("More than one value in iterable.")

        return sequence[0]

    def _getTemperatureVariableNameForDirectPortPairPortItem(self, directPortPair, portItem):
        isInputPort = directPortPair.fromPort == portItem
        relativeHeightInPercent = (
            directPortPair.relativeInputHeightPercent if isInputPort else directPortPair.relativeOutputHeightPercent
        )
        return f"T{self.displayName}Port{directPortPair.side.formatDdck()}{relativeHeightInPercent}"

    @staticmethod
    def _getTemperatureVariableNameForHeatExchangerPortItem(heatExchanger):
        return f"T{heatExchanger.displayName}"

    # Misc
    def contextMenuEvent(self, event):
        menu = QMenu()

        launchNotepadAction = menu.addAction("Launch NotePad++")
        launchNotepadAction.triggered.connect(self.launchNotepadFile)

        rotateRightIcon = _img.ROTATE_TO_RIGHT_PNG.icon()
        rotateRightAction = menu.addAction(rotateRightIcon, "Rotate Block clockwise")
        rotateRightAction.triggered.connect(self.rotateBlockCW)

        rotateLeftIcon = _img.ROTATE_LEFT_PNG.icon()
        rotateLeftIcon = menu.addAction(rotateLeftIcon, "Rotate Block counter-clockwise")
        rotateLeftIcon.triggered.connect(self.rotateBlockCCW)

        resetRotationAction = menu.addAction("Reset Rotation")
        resetRotationAction.triggered.connect(self.resetRotation)

        printRotationAction = menu.addAction("Print Rotation")
        printRotationAction.triggered.connect(self.printRotation)

        deleteBlockAction = menu.addAction("Delete this Block")
        deleteBlockAction.triggered.connect(self.deleteBlockCom)

        exportDdckAction = menu.addAction("Export ddck")
        exportDdckAction.triggered.connect(self.exportDck)

        menu.exec(event.screenPos())

    def mouseDoubleClickEvent(self, event):
        ConfigureStorageDialog(self, self.scene().parent())

    # Export related
    def exportBlackBox(self):
        equations = []
        ddcxPath = _os.path.join(self.path, self.displayName)
        ddcxPath = ddcxPath + ".ddcx"
        self.exportDck()
        if _os.path.isfile(ddcxPath):
            with open(ddcxPath, "r", encoding="windows-1252") as infile:
                lines = infile.readlines()
            for line in lines:
                if line[0] == "T":
                    equations.append(line.replace("\n", ""))
            return "success", equations

        self.logger.warning("No file at " + ddcxPath)
        return "noDdckFile", equations

    def getInternalPiping(self) -> InternalPiping:
        heatExchangerNodes, heatExchangerPortItems = self._createHeatExchangerNodes()

        portPairNodes, portPairsPortItems = self._createPortPairNodes()

        nodes = [*heatExchangerNodes, *portPairNodes]
        modelPortItemsToGraphicalPortItem = heatExchangerPortItems | portPairsPortItems

        return InternalPiping(nodes, modelPortItemsToGraphicalPortItem)

    def _createHeatExchangerNodes(self):
        heatExchangerPortItems = {}
        heatExchangerNodes = []
        for heatExchanger in self.heatExchangers:
            heatExchangerPortItem1 = _mfn.PortItem("Heat Exchanger Input", _mfn.PortItemType.INPUT)
            heatExchangerPortItems[heatExchangerPortItem1] = heatExchanger.port1

            heatExchangerPortItem2 = _mfn.PortItem("Heat Exchanger Output", _mfn.PortItemType.OUTPUT)
            heatExchangerPortItems[heatExchangerPortItem2] = heatExchanger.port2

            name = self._getMassFlowVariableSuffixForHeatExchanger(heatExchanger)

            node = _mfn.Pipe(name, heatExchanger.trnsysId, heatExchangerPortItem1, heatExchangerPortItem2)
            heatExchangerNodes.append(node)

        return heatExchangerNodes, heatExchangerPortItems

    def _createPortPairNodes(self):
        portPairsPortItems = {}
        portPairNodes = []
        for directPortPair in self.directPortPairs:
            portPairPortItem1 = _mfn.PortItem("Input", _mfn.PortItemType.INPUT)
            portPairsPortItems[portPairPortItem1] = directPortPair.fromPort

            portPairPortItem2 = _mfn.PortItem("Output", _mfn.PortItemType.OUTPUT)
            portPairsPortItems[portPairPortItem2] = directPortPair.toPort

            portPairName = self._getMassFlowVariableSuffixForDirectPortPair(directPortPair)

            node = _mfn.Pipe(portPairName, directPortPair.trnsysId, portPairPortItem1, portPairPortItem2)
            portPairNodes.append(node)

        return portPairNodes, portPairsPortItems

    def _getMassFlowVariableSuffixForDirectPortPair(self, directPortPair: DirectPortPair):
        return (
            f"{self.displayName}Dp{'L' if directPortPair.side.isLeft else 'R'}"
            f"{directPortPair.relativeInputHeightPercent}-{directPortPair.relativeOutputHeightPercent}"
        )

    @staticmethod
    def _getMassFlowVariableSuffixForHeatExchanger(heatExchanger):
        return heatExchanger.displayName

    def exportDck(self):  # pylint: disable=too-many-locals,too-many-statements

        if not self._checkConnExists():
            msgb = QMessageBox()
            msgb.setText("Please connect all ports before exporting!")
            msgb.exec_()
            return
        noError = self._debugConn()

        if not noError:
            qmb = QMessageBox()
            qmb.setText("Ignore connection errors and continue with export?")
            qmb.setStandardButtons(QMessageBox.Save | QMessageBox.Cancel)
            qmb.setDefaultButton(QMessageBox.Cancel)
            ret = qmb.exec()
            if ret == QMessageBox.Save:
                self.logger.debug("Overwriting")
            else:
                self.logger.debug("Canceling")
                return

        nPorts = len(self.directPortPairs)
        nHx = len(self.heatExchangers)

        tool = Type1924_TesPlugFlow()

        inputs = {
            "nUnit": 50,
            "nType": self.storageType,
            "nTes": self.nTes,
            "nPorts": nPorts,
            "nHx": nHx,
            "nHeatSources": 1,
        }

        directPairsPorts = []
        for directPortPair in self.directPortPairs:
            incomingConnection = directPortPair.fromPort.connectionList[0]
            temperatureName = "T" + incomingConnection.displayName
            massFlowRateName = "Mfr" + incomingConnection.displayName

            outgoingConnection = directPortPair.toPort.connectionList[0]
            reverseTemperatureName = "T" + outgoingConnection.displayName

            inputPos = directPortPair.relativeInputHeight
            outputPos = directPortPair.relativeOutputHeight

            directPairsPort = {
                "T": temperatureName,
                "side": directPortPair.side.formatDdck(),
                "Mfr": massFlowRateName,
                "Trev": reverseTemperatureName,
                "zIn": inputPos,
                "zOut": outputPos,
            }
            directPairsPorts.append(directPairsPort)

        heatExchangerPorts = []
        for heatExchanger in self.heatExchangers:
            heatExchangerName = heatExchanger.displayName
            incomingConnection = heatExchanger.port1.connectionList[0]
            temperatureName = "T" + incomingConnection.displayName
            massFlowRateName = "Mfr" + incomingConnection.displayName

            outgoingConnection = heatExchanger.port2.connectionList[0]
            reverseTemperatureName = "T" + outgoingConnection.displayName

            inputPos = heatExchanger.relativeInputHeight
            outputPos = heatExchanger.relativeOutputHeight

            heatExchangerPort = {
                "Name": heatExchangerName,
                "T": temperatureName,
                "Mfr": massFlowRateName,
                "Trev": reverseTemperatureName,
                "zIn": inputPos,
                "zOut": outputPos,
                "cp": "cpwat",
                "rho": "rhowat",
            }

            heatExchangerPorts.append(heatExchangerPort)

        auxiliaryPorts = []
        for _ in range(inputs["nHeatSources"]):
            dictInputAux = {"zAux": 0.0, "qAux": 0.0}
            auxiliaryPorts.append(dictInputAux)

        tool.setInputs(inputs, directPairsPorts, heatExchangerPorts, auxiliaryPorts)

        tool.createDDck(self.path, self.displayName, typeFile="ddck")

    def _debugConn(self):
        self.logger.debug("Debugging conn")
        errorConnList = ""
        for directPort in self.directPortPairs:
            stFromPort = directPort.fromPort
            stToPort = directPort.toPort
            toPort1 = stFromPort.connectionList[0].toPort
            fromPort2 = stToPort.connectionList[0].fromPort
            connName1 = stFromPort.connectionList[0].displayName
            connName2 = stToPort.connectionList[0].displayName

            if stFromPort != toPort1:
                errorConnList = errorConnList + connName1 + "\n"
            if stToPort != fromPort2:
                errorConnList = errorConnList + connName2 + "\n"
        if errorConnList != "":
            msgBox = QMessageBox()
            msgBox.setText(f"{errorConnList} is connected wrongly, right click StorageTank to invert connection.")
            msgBox.exec()
            noError = False
        else:
            noError = True

        return noError

    def _checkConnExists(self):
        for heatExchanger in self.heatExchangers:
            if not heatExchanger.port1.connectionList:
                return False
            if not heatExchanger.port2.connectionList:
                return False

        for ports in self.leftDirectPortPairsPortItems:
            if not ports.connectionList:
                return False

        for ports in self.rightDirectPortPairsPortItems:
            if not ports.connectionList:
                return False

        return True

    def addTree(self):
        """
        When a blockitem is added to the main window.
        A file explorer for that item is added to the right of the main window by calling this method
        """
        pathName = self.displayName
        if self.parent.parent().projectPath == "":
            self.path = self.parent.parent().projectFolder
        else:
            self.path = self.parent.parent().projectPath
        self.path = _os.path.join(self.path, "ddck")
        self.path = _os.path.join(self.path, pathName)
        if not _os.path.exists(self.path):
            _os.makedirs(self.path)

        self.model = MyQFileSystemModel()
        self.model.setRootPath(self.path)
        self.model.setName(self.displayName)
        self.tree = MyQTreeView(self.model, self)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(self.path))
        self.tree.setObjectName(f"{self.displayName}Tree")
        self.tree.setMinimumHeight(200)
        self.tree.setSortingEnabled(True)
        self.parent.parent().splitter.addWidget(self.tree)

    def deleteBlock(self):
        """
        Overridden method to also delete folder
        """
        self.logger.debug("Block " + str(self) + " is deleting itself (" + self.displayName + ")")
        self.deleteConns()
        self.parent.parent().trnsysObj.remove(self)
        self.logger.debug("deleting block " + str(self) + self.displayName)
        self.parent.scene().removeItem(self)
        widgetToRemove = self.parent.parent().findChild(QTreeView, self.displayName + "Tree")
        _sh.rmtree(self.path)
        try:
            widgetToRemove.hide()
        except AttributeError:
            self.logger.debug("Widget doesnt exist!")
        else:
            self.logger.debug("Deleted widget")
        del self

    def setDisplayName(self, newName):
        """
        Overridden method to also change folder name
        """
        self.displayName = newName
        self.label.setPlainText(newName)
        self.model.setName(self.displayName)
        self.tree.setObjectName(f"{self.displayName}Tree")
        self.logger.debug(_os.path.dirname(self.path))
        destPath = _os.path.join(_os.path.split(self.path)[0], self.displayName)
        if _os.path.split(self.path)[-1] == "" or _os.path.split(self.path)[-1] == "ddck":
            _os.makedirs(destPath)
        else:
            if _os.path.exists(self.path):
                _os.rename(self.path, destPath)
        self.path = destPath
        self.logger.debug(self.path)



import PyQt5.QtWidgets as _qtw
from PyQt5.QtCore import QPointF, Qt
from PyQt5.QtGui import QPen

from trnsysGUI.BlockItem import BlockItem  # type: ignore[attr-defined]
from trnsysGUI.Graphicaltem import GraphicalItem  # type: ignore[attr-defined]
from trnsysGUI.ResizerItem import ResizerItem  # type: ignore[attr-defined]
from trnsysGUI.connection.connectionBase import ConnectionBase  # type: ignore[attr-defined]
from trnsysGUI.storageTank.widget import StorageTank


class Scene(_qtw.QGraphicsScene):
    """
    This class serves as container for QGraphicsItems and is used in combination with the View to display the
    diagram.
    """

    def __init__(self, editor):
        super().__init__(editor)

        self.logger = editor.logger
        self._editor = editor

        self._previouslyHitItems = []

    def mouseMoveEvent(self, mouseEvent):
        self.parent().sceneMouseMoveEvent(mouseEvent)
        super().mouseMoveEvent(mouseEvent)

    def mouseReleaseEvent(self, mouseEvent):
        self.parent().sceneMouseReleaseEvent(mouseEvent)
        super().mouseReleaseEvent(mouseEvent)
        self.parent().moveDirectPorts = False

    def drawBackground(self, painter, rect):
        # Overwrite drawBackground if snapGrid is True
        if self.parent().snapGrid:
            pen = QPen()
            pen.setWidth(2)
            pen.setCosmetic(True)
            painter.setPen(pen)

            gridSize = self.parent().snapSize

            left = int(rect.left()) - (int(rect.left()) % gridSize)
            top = int(rect.top()) - (int(rect.top()) % gridSize)
            points = []
            for xIndex in range(left, int(rect.height()), gridSize):
                for yIndex in range(top, int(rect.bottom()), gridSize):
                    points.append(QPointF(xIndex, yIndex))

            for point in points:
                painter.drawPoint(point)
        else:
            super().drawBackground(painter, rect)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_L:
            self.parent().moveDirectPorts = not self.parent().moveDirectPorts
            return

        if event.key() == Qt.Key_Delete:
            trnsysObjects = self.parent().trnsysObj
            selectedObjects = [o for o in trnsysObjects if o.isSelected]

            if not selectedObjects:
                return

            assert len(selectedObjects) == 1, "It shouldn't be possible select more than one object"

            selectedObject = selectedObjects[0]

            if isinstance(selectedObject, ConnectionBase):
                selectedObject.createDeleteUndoCommandAndAddToStack()
            if isinstance(selectedObject, BlockItem):
                selectedObject.deleteBlockCom()

    def mousePressEvent(self, event):
        if not self._editor.parent().massFlowEnabled:
            for connection in self._editor.connectionList:
                connection.deselectConnection()

        for trnsysObject in self._editor.trnsysObj:
            if isinstance(trnsysObject, BlockItem):
                trnsysObject.isSelected = False
            if isinstance(trnsysObject, StorageTank):
                for heatExchanger in trnsysObject.heatExchangers:
                    heatExchanger.unhighlightHx()

        self.parent().alignYLineItem.setVisible(False)

        if self._previouslyHitItems is not None:
            for item in self._previouslyHitItems:
                if isinstance(item, (GraphicalItem, BlockItem)) and hasattr(item, "resizer"):
                    self.removeItem(item.resizer)
                    item.deleteResizer()
                if isinstance(item, ResizerItem):
                    self.removeItem(item)
                    item.parent.deleteResizer()

            self._previouslyHitItems.clear()

        super().mousePressEvent(event)

        hitItems = self.items(event.scenePos())
        self._previouslyHitItems = hitItems



import typing as _tp

import trnsysGUI.images as _img
from trnsysGUI.sourceSinkBase import SourceSinkBase


class Water(SourceSinkBase):
    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.WATER_SVG



# pylint: skip-file
import dataclasses as _dc
import logging as _log
import re
import typing as _tp

import jinja2 as _jinja

import trnsysGUI.TVentil as _tventil
import trnsysGUI.connection.connectionBase as _cb
import trnsysGUI.connection.doublePipeConnection as _dpc
import trnsysGUI.connection.singlePipeConnection as _spc
import trnsysGUI.connection.values as _values
import trnsysGUI.hydraulicLoops.model as _hlm
import trnsysGUI.hydraulicLoops.names as _names
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.globalNetwork as _gn
import trnsysGUI.massFlowSolver.networkModel as _mfn

_UNUSED_INDEX = 0


@_dc.dataclass
class _SerializedNode:
    name: str
    index: int
    nodeType: int
    neighborIndexes: _tp.Tuple[int, int, int]


class Export:
    def __init__(
        self,
        diagramName: str,
        massFlowContributors: _tp.Sequence[_mfs.MassFlowNetworkContributorMixin],
        hydraulicLoops: _tp.Sequence[_hlm.HydraulicLoop],
        fluids: _tp.Sequence[_hlm.Fluid],
        logger: _log.Logger,
        editor,
    ) -> None:
        self._hydraulicLoops = hydraulicLoops
        self._fluids = fluids
        self._diagramName = diagramName
        self._massFlowContributors = massFlowContributors
        self.logger = logger
        self._editor = editor

        self.maxChar = 20

        o: _mfs.MassFlowNetworkContributorMixin
        nodes = [n for c in self._massFlowContributors for n in c.getInternalPiping().openLoopsStartingNodes]
        self.lineNumOfPar = len(nodes)

        self.numOfPar = 4 * self.lineNumOfPar + 1

    def exportBlackBox(self, exportTo="ddck"):
        f = "*** Black box component temperatures" + "\n"
        equationNr = 0
        problemEncountered = False

        for t in self._massFlowContributors:
            status, equations = t.exportBlackBox()

            for equation in equations:
                f += equation + "\n"
            equationNr += len(equations)

        if exportTo == "mfs":
            lines = f.split("\n")
            f = ""
            for i in range(len(lines)):
                if "=" in lines[i]:
                    lines[i] = lines[i].split("=")[0] + "=1"
                f += lines[i] + "\n"

        if equationNr == 0:
            f = ""
        else:
            f = "\nEQUATIONS " + str(equationNr) + "\n" + f + "\n"

        return problemEncountered, f

    def exportDoublePipeParameters(self, exportTo="ddck"):
        doesHydraulicContainDoublePipes = any(
            [isinstance(obj, _dpc.DoublePipeConnection) for obj in self._massFlowContributors]
        )

        if not doesHydraulicContainDoublePipes:
            return ""

        commentStars = 6 * "*"

        if exportTo == "ddck":
            exportText = "CONSTANTS 25\n"
        elif exportTo == "mfs":
            exportText = "CONSTANTS 28\n"
        else:
            raise ValueError("Unknown value for `exportTo`", exportTo)

        exportText += "*** Default global PARAMETERS for TYPES 9511" + "\n"

        exportText += commentStars + " pipe and soil properties " + commentStars + "\n"
        exportText += self._addComment("dpLength = 24.384", "! Length of buried pipe, m")
        exportText += self._addComment("dpDiamIn = 0.02618", "! Inner diameter of pipes, m")
        exportText += self._addComment("dpDiamOut = 0.03198", "! Outer diameter of pipes, m")
        exportText += self._addComment("dpLambda = 1.37067", "! Thermal conductivity of pipe material, kJ/(h*m*K)")
        exportText += self._addComment("dpDepth = 3.0", "! Buried pipe depth, m")
        exportText += self._addComment("dpFlowMode = 2", "! Direction of second pipe flow: 1 = same, 2 = opposite")
        exportText += self._addComment("dpDiamCase = 0.17526", "! Diameter of casing material, m")
        exportText += self._addComment("dpLambdaFill = 0.14537", "! Thermal conductivity of fill insulation, kJ/(h*m*K)")
        exportText += self._addComment("dpDistPtoP = 0.06911", "! Center-to-center pipe spacing, m")
        exportText += self._addComment("dpLambdaGap = 8.722", "! Thermal conductivity of gap material, kJ/(h*m*K)")
        exportText += self._addComment("dpGapThick = 0.0000", "! Gap thickness, m")

        exportText += commentStars + " fluid properties " + commentStars + "\n"
        exportText += self._addComment("dpRhoFlu = 1000.0", "! Density of fluid, kg/m^3")
        exportText += self._addComment("dpLambdaFl = 2.2068", "! Thermal conductivity of fluid, kJ/(h*m*K)")
        exportText += self._addComment("dpCpFl = 4.19", "! Specific heat of fluid, kJ/(kg*K)")
        exportText += self._addComment("dpViscFl = 3.078", "! Viscosity of fluid, kg/(m*h)")

        exportText += commentStars + " initial conditions " + commentStars + "\n"
        exportText += self._addComment("dpTIniHot = 10.0", "! Initial fluid temperature - Pipe hot, deg C")
        exportText += self._addComment("dpTIniCold = 10.0", "! Initial fluid temperature - Pipe cold, deg C")

        exportText += commentStars + " thermal properties soil " + commentStars + "\n"
        exportText += self._addComment("dpLamdaSl = 8.722", "! Thermal conductivity of soil, kJ/(h*m*K)")
        exportText += self._addComment("dpRhoSl = 2500.0", "! Density of soil, kg/m^3")
        exportText += self._addComment("dpCpSl = 0.84", "! Specific heat of soil, kJ/(kg*K)")

        if exportTo == "mfs":
            exportText += (
                commentStars + " general temperature dependency (dependent on weather data) " + commentStars + "\n"
            )
            exportText += self._addComment("TambAvg = 7.96", "! Average surface temperature, deg C")
            exportText += self._addComment("dTambAmpl = 13.32", "! Amplitude of surface temperature, deg C")
            exportText += self._addComment("ddTcwOffset = 36", "! Days of minimum surface temperature")

        exportText += commentStars + " definition of nodes " + commentStars + "\n"
        exportText += self._addComment("dpNrFlNds = 100", "! Number of fluid nodes")
        exportText += self._addComment("dpNrSlRad = 8", "! Number of radial soil nodes")
        exportText += self._addComment("dpNrSlAx = 10", "! Number of axial soil nodes")
        exportText += self._addComment("dpNrSlCirc = 4", "! Number of circumferential soil nodes")
        exportText += self._addComment("dpRadNdDist = 0.0254", "! Radial distance of any node, m")

        exportText += "\n"

        return exportText

    def exportPumpOutlets(self):
        f = "*** Pump outlet temperatures" + "\n"
        equationNr = 0
        for t in self._massFlowContributors:
            f += t.exportPumpOutlets()[0]
            equationNr += t.exportPumpOutlets()[1]

        if equationNr == 0:
            f = ""
        else:
            f = "EQUATIONS " + str(equationNr) + "\n" + f + "\n"

        return f

    def exportMassFlows(self):  # What the controller should give
        f = "*** Massflowrates" + "\n"
        equationNr = 0

        for t in self._massFlowContributors:
            f += t.exportMassFlows()[0]
            equationNr += t.exportMassFlows()[1]

        if equationNr == 0:
            f = ""
        else:
            f = "EQUATIONS " + str(equationNr) + "\n" + f + "\n"

        return f

    def exportDivSetting(self, unit):
        """
        :param unit: the index of the previous unit number used.
        :return:
        """
        f = ""

        nUnit = unit
        constants = 0
        f2 = ""
        for t in self._massFlowContributors:
            f2 += t.exportDivSetting1()[0]
            constants += t.exportDivSetting1()[1]

        if constants > 0:
            f = "CONSTANTS " + str(constants) + "\n"
            f += f2 + "\n"

        for t in self._massFlowContributors:
            res = t.exportDivSetting2(nUnit)
            f += res[0]
            nUnit = res[1]

        return f

    def exportParametersFlowSolver(self, simulationUnit: int, simulationType: int, descConnLength: int) -> str:
        massFlowContributors = []
        for massFlowContributor in self._massFlowContributors:
            noHydraulicConnection = (
                not isinstance(massFlowContributor, _cb.ConnectionBase)
                and not massFlowContributor.outputs  # type: ignore[attr-defined]
                and not massFlowContributor.inputs  # type: ignore[attr-defined]
            )

            if noHydraulicConnection:
                continue

            massFlowContributors.append(massFlowContributor)

        serializedNodes = self._getSerializedNodes(massFlowContributors)

        lines = []
        for serializedNode in serializedNodes:
            neighborIndexes = serializedNode.neighborIndexes

            firstColumn = f"{neighborIndexes[0]} {neighborIndexes[1]} {neighborIndexes[2]} {serializedNode.nodeType} "
            secondColumn = f"!{serializedNode.index} : {serializedNode.name}"

            line = f"{firstColumn.ljust(descConnLength)}{secondColumn}"

            lines.append(line)

        jointLines = "\n".join(lines)

        resultText = f"""\
UNIT {simulationUnit} TYPE {simulationType}
PARAMETERS {len(serializedNodes) * 4 + 1}
{len(serializedNodes)}
{jointLines}
"""

        return resultText

    @classmethod
    def _getSerializedNodes(
        cls,
        massFlowContributors: _tp.Sequence[_mfs.MassFlowNetworkContributorMixin],
    ) -> _tp.Sequence[_SerializedNode]:
        globalNetwork = _gn.getGlobalNetwork(massFlowContributors)
        internalPortItemToExternalRealNode = globalNetwork.internalPortItemToExternalRealNode

        realNodesToIndex = {r: i for i, r in enumerate(globalNetwork.realNodes, start=1)}

        serializedNodes = []
        for index, realNode in enumerate(globalNetwork.realNodes, start=1):
            neighborsAndUnusedIndexes = cls._getNeighborAndUnusedIndexes(
                realNode, realNodesToIndex, internalPortItemToExternalRealNode
            )

            serializedNode = _SerializedNode(realNode.name, index, realNode.getNodeType(), neighborsAndUnusedIndexes)

            serializedNodes.append(serializedNode)

        return serializedNodes

    @staticmethod
    def _getNeighborAndUnusedIndexes(
        realNode: _mfn.RealNodeBase,
        realNodesToIndex: _tp.Mapping[_mfn.RealNodeBase, int],
        internalPortItemToExternalRealNode: _tp.Mapping[_mfn.PortItem, _mfn.RealNodeBase],
    ) -> _tp.Tuple[int, int, int]:
        neighborIndexes = []
        for neighbor in realNode.getNeighbours():
            if isinstance(neighbor, _mfn.RealNodeBase):
                neighborIndex = realNodesToIndex[neighbor]
            elif isinstance(neighbor, _mfn.PortItem):
                externalRealNode = internalPortItemToExternalRealNode[neighbor]
                neighborIndex = realNodesToIndex[externalRealNode]
            else:
                raise AssertionError("Can't get here.")

            neighborIndexes.append(neighborIndex)

        nNeighborIndexes = len(neighborIndexes)

        assert nNeighborIndexes > 0
        neighborAndUnusedIndexes = [neighborIndexes[0], _UNUSED_INDEX, _UNUSED_INDEX]

        if nNeighborIndexes > 1:
            neighborAndUnusedIndexes[1] = neighborIndexes[1]

        if nNeighborIndexes > 2:
            neighborAndUnusedIndexes[2] = neighborIndexes[2]

        return neighborAndUnusedIndexes[0], neighborAndUnusedIndexes[1], neighborAndUnusedIndexes[2]

    def convertToStringList(self, l):
        res = []
        temp = ""
        for i in l:
            if i == "\n":
                res.append(temp)
                temp = ""
            else:
                temp += i

        return res

    def findId(self, s):
        a = s[s.find("!") + 1 : s.find(" ", s.find("!"))]
        self.logger.debug(a)
        return a

    def _correctIds(self, lineList):
        fileCopy = lineList[:]
        self.logger.debug("fds" + str(fileCopy))
        fileCopy = [" " + l for l in fileCopy]

        matchNumber = re.compile(r"\d+")

        counter = 0
        for line in lineList:
            counter += 1
            k = self.findId(line)
            if k != str(counter):

                descConnlen = 15
                res = fileCopy[:]
                for l in range(len(fileCopy)):
                    ids = matchNumber.findall(fileCopy[l])
                    for i in range(3):
                        if ids[i] == str(k):
                            ids[i] = str(counter)

                    fileCopyTempLine = fileCopy[l]
                    rest = fileCopyTempLine[fileCopyTempLine.find("!") :]
                    res[l] = ids[0] + " " + ids[1] + " " + ids[2] + " " + ids[3]
                    res[l] += " " * (descConnlen - len(res[l])) + rest

                fileCopy = res
                fileCopy = [l.replace("!" + str(k) + " ", "!" + str(counter) + " ") for l in fileCopy]

        return "\n".join(fileCopy)

    def exportInputsFlowSolver(self):
        f = ""
        f += "INPUTS " + str(self.lineNumOfPar) + "! for Type 935\n"

        numberOfInputs = 0

        counter = 0
        for t in self._massFlowContributors:
            res = t.exportInputsFlowSolver()
            f += res[0]
            counter += res[
                1
            ]  # DC this is a very strange way to print values, I would like to have 10 values per line and the fact that two go together makes it complicated
            numberOfInputs += res[1]

            if counter > 9 or t == self._massFlowContributors[-1]:
                f += "\n"
                counter = 0

        f += "*** Initial Inputs\n"
        counter = 0
        for _ in range(numberOfInputs):
            f += "0 "
            counter += 1
            # DC if we use 10 per line its way easier to count and see if it is well done.
            # Above should be the same but some elements give two at once and depending if they are at then end of the line 11 values show up
            if counter > 9:
                f += "\n"
                counter = 0

        f += "\n\n"

        return f

    def exportOutputsFlowSolver(self, simulationUnit):
        f = ""

        abc = "ABC"

        prefix = "Mfr"
        equationNumber = 1
        nEqUsed = 1  # DC

        tot = ""

        for t in self._massFlowContributors:
            noHydraulicConnection = not isinstance(t, _cb.ConnectionBase) and not t.outputs and not t.inputs

            if noHydraulicConnection:
                continue
            else:
                res = t.exportOutputsFlowSolver(prefix, abc, equationNumber, simulationUnit)
                tot += res[0]
                equationNumber = res[1]
                nEqUsed += res[2]

        head = (
            "EQUATIONS {0}	! Output up to three (A,B,C) mass flow rates of each component, positive = "
            "input/inlet, negative = output/outlet ".format(nEqUsed - 1)
        )

        f += head + "\n"
        f += tot + "\n"
        f += "\n"

        return f

    def exportPipeAndTeeTypesForTemp(self, startingUnit):
        # Prints the part of the export where the pipes, tp and div Units are printed

        f = ""
        unitNumber = startingUnit
        # typeNr1 = 929 # Temperature calculation from a tee-piece
        # typeNr2 = 931 # Temperature calculation from a pipe

        for t in self._massFlowContributors:

            res = t.exportPipeAndTeeTypesForTemp(unitNumber)
            f += res[0]
            unitNumber = res[1]

        self._editor.printerUnitnr = unitNumber

        return f

    def exportPrintPipeLosses(self):
        f = ""
        lossText = ""
        rightCounter = 0

        for t in self._massFlowContributors:
            if isinstance(t, _spc.SinglePipeConnection):
                if rightCounter != 0:
                    lossText += "+"
                lossText += "P" + t.displayName + "_kW"
                rightCounter += 1
            if isinstance(t, _dpc.DoublePipeConnection):
                if rightCounter != 0:
                    lossText += "+"
                lossText += "P" + t.displayName + "Cold_kW" + "+"
                lossText += "P" + t.displayName + "Hot_kW"
                rightCounter += 1

        if rightCounter == 0:
            lossText += "0"

        f += "*** Pipe losses\nEQUATIONS 1\nPipeLossTot=" + lossText + "\n\n"
        return f

    def exportMassFlowPrinter(self, unitnr, descLen):
        typenr = 25
        printingMode = 0
        relAbsStart = 0
        overwriteApp = -1
        printHeader = -1
        delimiter = 0
        printLabels = 1

        f = "ASSIGN " + self._diagramName.rsplit(".", 1)[0] + "_Mfr.prt " + str(unitnr) + "\n\n"

        f += "UNIT " + str(unitnr) + " TYPE " + str(typenr)
        f += " " * (descLen - len(f)) + "! User defined Printer" + "\n"
        f += "PARAMETERS 10" + "\n"
        f += "dtSim"
        f += " " * (descLen - len(f)) + "! 1 Printing interval" + "\n"
        f += "START"
        f += " " * (descLen - len(f)) + "! 2 Start time" + "\n"
        f += "STOP"
        f += " " * (descLen - len(f)) + "! 3 Stop time" + "\n"
        f += str(unitnr)
        f += " " * (descLen - len(f)) + "! 4 Logical unit" + "\n"
        f += str(printingMode)
        f += " " * (descLen - len(f)) + "! 5 Units printing mode" + "\n"
        f += str(relAbsStart)
        f += " " * (descLen - len(f)) + "! 6 Relative or absolute start time" + "\n"
        f += str(overwriteApp)
        f += " " * (descLen - len(f)) + "! 7 Overwrite or Append" + "\n"
        f += str(printHeader)
        f += " " * (descLen - len(f)) + "! 8 Print header" + "\n"
        f += str(delimiter)
        f += " " * (descLen - len(f)) + "! 9 Delimiter" + "\n"
        f += str(printLabels)
        f += " " * (descLen - len(f)) + "! 10 Print labels" + "\n"
        f += "\n"

        s = ""
        equationType = "Mfr"
        breakline = 0
        for t in self._massFlowContributors:
            if isinstance(t, _spc.SinglePipeConnection):
                breakline, s = self._getEquation(breakline, s, t, "", equationType)
            if isinstance(t, _dpc.DoublePipeConnection):
                breakline, s = self._getEquation(breakline, s, t, "Cold", equationType)
                breakline, s = self._getEquation(breakline, s, t, "Hot", equationType)
            if isinstance(t, _tventil.TVentil) and t.isVisible():
                breakline += 1
                if breakline % 8 == 0:
                    s += "\n"
                s += "xFrac" + t.displayName + " "
        f += "INPUTS " + str(breakline) + "\n" + s + "\n" + "***" + "\n" + s + "\n\n"

        return f

    def _getEquation(self, breakline, s, t, temperature, equationType):
        breakline += 1
        if breakline % 8 == 0:
            s += "\n"
        s += equationType + t.displayName + temperature + " "
        return breakline, s

    @staticmethod
    def _addComment(firstColumn, comment):
        spacing = 40
        return str(firstColumn).ljust(spacing) + comment + "\n"

    def exportTempPrinter(self, unitnr, descLen):

        typenr = 25
        printingMode = 0
        relAbsStart = 0
        overwriteApp = -1
        printHeader = -1
        delimiter = 0
        printLabels = 1

        f = "ASSIGN " + self._diagramName.rsplit(".", 1)[0] + "_T.prt " + str(unitnr) + "\n\n"

        f += "UNIT " + str(unitnr) + " TYPE " + str(typenr)
        f += " " * (descLen - len(f)) + "! User defined Printer" + "\n"
        f += "PARAMETERS 10" + "\n"
        f += "dtSim"
        f += " " * (descLen - len(f)) + "! 1 Printing interval" + "\n"
        f += "START"
        f += " " * (descLen - len(f)) + "! 2 Start time" + "\n"
        f += "STOP"
        f += " " * (descLen - len(f)) + "! 3 Stop time" + "\n"
        f += str(unitnr)
        f += " " * (descLen - len(f)) + "! 4 Logical unit" + "\n"
        f += str(printingMode)
        f += " " * (descLen - len(f)) + "! 5 Units printing mode" + "\n"
        f += str(relAbsStart)
        f += " " * (descLen - len(f)) + "! 6 Relative or absolute start time" + "\n"
        f += str(overwriteApp)
        f += " " * (descLen - len(f)) + "! 7 Overwrite or Append" + "\n"
        f += str(printHeader)
        f += " " * (descLen - len(f)) + "! 8 Print header" + "\n"
        f += str(delimiter)
        f += " " * (descLen - len(f)) + "! 9 Delimiter" + "\n"
        f += str(printLabels)
        f += " " * (descLen - len(f)) + "! 10 Print labels" + "\n"
        f += "\n"

        s = ""
        equationType = "T"
        breakline = 0
        for t in self._massFlowContributors:
            if isinstance(t, _spc.SinglePipeConnection):
                breakline, s = self._getEquation(breakline, s, t, "", equationType)
            if isinstance(t, _dpc.DoublePipeConnection):
                breakline, s = self._getEquation(breakline, s, t, "Cold", equationType)
                breakline, s = self._getEquation(breakline, s, t, "Hot", equationType)
        f += "INPUTS " + str(breakline) + "\n" + s + "\n" + "***" + "\n" + s + "\n\n"

        return f

    def exportFluids(self) -> str:
        def getValueOrVariableName(
            valueOrVariable: _tp.Union[float, _hlm.Variable], factor: float = 1
        ) -> _tp.Union[float, str]:
            if isinstance(valueOrVariable, float):
                return valueOrVariable * factor

            if isinstance(valueOrVariable, _hlm.Variable):
                if factor == 1:
                    return valueOrVariable.name

                return f"{valueOrVariable.name}*{factor}"

            raise ValueError(f"Can't deal with type {type(valueOrVariable).__name__}.")

        template = """\
** Fluids:
EQUATIONS {{2 * fluids|length}}
{% for fluid in fluids -%}
** {{fluid.name}}
{% set rho = getValueOrVariableName(fluid.densityInKgPerM3) -%}
{% set cp = getValueOrVariableName(fluid.specificHeatCapacityInJPerKgK, 1e-3) -%}
F{{fluid.name}}Rho = {{rho}} ! [kg/m^3]
F{{fluid.name}}Cp = {{cp}} ! [kJ/(kg*K)]
{% endfor -%}
"""
        fluids = self._fluids
        return self._render(template, fluids=fluids, getValueOrVariableName=getValueOrVariableName)

    def exportHydraulicLoops(self) -> str:
        template = """\
** Hydraulic loops
EQUATIONS {{nEquations}}
{% for hydraulicLoop in loops -%}
{% set loopName = hydraulicLoop.name.value -%}
{% set loopRho = names.getDensityName(loopName) -%}
{% set loopCp = names.getHeatCapacityName(loopName) -%}
{% set fluid = hydraulicLoop.fluid -%}
{% set loopLen = names.getDefaultLengthName(loopName) -%}
{% set loopDia = names.getDefaultDiameterName(loopName) -%}
{% set loopUVal = names.getDefaultUValueName(loopName) -%}
{% set loopNPipes = names.getNumberOfPipesName(loopName) -%}
** {{loopName}}
{% if hydraulicLoop.useLoopWideDefaults -%}
{{loopNPipes}} = {{hydraulicLoop.connections | length}}
{{loopLen}} = {{values.DEFAULT_LENGTH_IN_M}} ! [m]
{{loopDia}} = {{values.DEFAULT_DIAMETER_IN_CM / 100}} ! [m]
{{loopUVal}} = {{values.DEFAULT_U_VALUE_IN_W_PER_M2_K * 60*60/1000}} ! [kJ/(h*m^2*K)] (= {{values.DEFAULT_U_VALUE_IN_W_PER_M2_K}} W/(m^2*K))
{% endif -%}
{{loopRho}} = F{{fluid.name}}Rho
{{loopCp}} = F{{fluid.name}}Cp

{% endfor -%}
"""
        loops = self._hydraulicLoops

        nEquations = sum(6 if l.useLoopWideDefaults else 2 for l in loops)

        return self._render(template, loops=loops, nEquations=nEquations, names=_names, values=_values)

    @staticmethod
    def _render(template: str, /, **kwargs):
        compiledTemplate = _jinja.Template(template)
        return compiledTemplate.render(**kwargs)



import typing as _tp

_T = _tp.TypeVar("_T")


def getSingle(iterable: _tp.Iterable[_T]) -> _T:
    """The `iterable` will be fully exhausted."""

    value = getSingleOrNone(iterable)

    if not value:
        raise ValueError("Expected one element but got none.")

    return value


def getSingleOrNone(iterable: _tp.Iterable[_T]) -> _tp.Optional[_T]:
    """The `iterable` will be fully exhausted."""

    values = [*iterable]

    if not values:
        return None

    if len(values) > 1:
        raise ValueError("Expected one element but got more than one.")

    return values[0]



# pylint: disable = invalid-name

import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import dataclasses_jsonschema as _dcj
from PyQt5.QtWidgets import QGraphicsTextItem

import pytrnsys.utils.serialization as _ser
import trnsysGUI.blockItemModel as _bim
import trnsysGUI.createSinglePipePortItem as _cspi  # pylint: disable=cyclic-import
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.massFlowSolver import MassFlowNetworkContributorMixin


class TVentil(BlockItem, MassFlowNetworkContributorMixin):  # pylint: disable = too-many-instance-attributes
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.h = 40
        self.w = 40
        self.isTempering = False
        self.positionForMassFlowSolver = 1.0
        self.posLabel = QGraphicsTextItem(str(self.positionForMassFlowSolver), self)
        self.posLabel.setVisible(False)

        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))
        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))

        self.changeSize()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.T_VENTIL_SVG

    def changeSize(self):
        width = self.w
        height = self.h

        delta = 20
        # Limit the block size:
        height = max(height, 20)
        width = max(width, 40)

        # center label:
        rect = self.label.boundingRect()
        labelWidth = rect.width()
        labelPosX = (width - labelWidth) / 2

        self.label.setPos(labelPosX, height - self.flippedV * (height + height / 2))
        self.posLabel.setPos(labelPosX + 5, -15)

        self.origInputsPos = [[width, delta]]
        self.origOutputsPos = [[0, delta], [delta, 0]]

        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])
        self.outputs[1].setPos(self.origOutputsPos[1][0], self.origOutputsPos[1][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4
        self.outputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.outputs[1].side = (self.rotationN + 1 + 2 * self.flippedV) % 4

        return width, height

    def rotateBlockCW(self):
        super().rotateBlockCW()
        # Rotate valve position label back so it will always stay horizontal
        self._updateRotation()

    def rotateBlockCCW(self):
        super().rotateBlockCCW()
        # Rotate valve position label back so it will always stay horizontal
        self._updateRotation()

    def resetRotation(self):
        super().resetRotation()
        # Rotate valve position label back so it will always stay horizontal
        self._updateRotation()

    def _updateRotation(self):
        self.posLabel.setRotation(-self.rotationN * 90)

    def setComplexDiv(self, b):  # pylint: disable = invalid-name
        self.isTempering = bool(b)

    def setPositionForMassFlowSolver(self, f):  # pylint: disable = invalid-name
        self.positionForMassFlowSolver = f

    def encode(self):
        portListInputs = []
        portListOutputs = []

        for inp in self.inputs:
            portListInputs.append(inp.id)
        for output in self.outputs:
            portListOutputs.append(output.id)

        blockPosition = (float(self.pos().x()), float(self.pos().y()))

        tVentilModel = TVentilModel(
            self.name,
            self.displayName,
            blockPosition,
            self.id,
            self.trnsysId,
            portListInputs,  # pylint: disable = duplicate-code # 2
            portListOutputs,
            self.flippedH,
            self.flippedV,
            self.rotationN,
        )

        dictName = "Block-"

        dct = tVentilModel.to_dict()

        dct["IsTempering"] = self.isTempering
        dct["PositionForMassFlowSolver"] = self.positionForMassFlowSolver
        return dictName, dct

    def decode(self, i, resBlockList):
        model = TVentilModel.from_dict(i)

        self.setDisplayName(model.BlockDisplayName)
        self.setPos(float(model.blockPosition[0]), float(model.blockPosition[1]))
        self.id = model.Id
        self.trnsysId = model.trnsysId

        self.inputs[0].id = model.portsIdsIn[0]
        self.outputs[0].id = model.portsIdsOut[0]
        self.outputs[1].id = model.portsIdsOut[1]

        self.updateFlipStateH(model.flippedH)
        self.updateFlipStateV(model.flippedV)
        self.rotateBlockToN(model.rotationN)

        resBlockList.append(self)

        if "IsTempering" not in i or "PositionForMassFlowSolver" not in i:
            self.logger.debug("Old version of diagram")
            self.positionForMassFlowSolver = 1.0
        else:
            self.isTempering = i["IsTempering"]
            self.positionForMassFlowSolver = i["PositionForMassFlowSolver"]

    def decodePaste(self, i, offset_x, offset_y, resConnList, resBlockList, **kwargs):
        super().decodePaste(i, offset_x, offset_y, resConnList, resBlockList, **kwargs)
        if "IsTempering" not in i or "PositionForMassFlowSolver" not in i:
            self.logger.debug("Old version of diagram")
            self.positionForMassFlowSolver = 1.0
        else:
            self.isTempering = i["IsTempering"]
            self.positionForMassFlowSolver = i["PositionForMassFlowSolver"]

    def exportMassFlows(self):
        if not self.isTempering:
            resStr = "xFrac" + self.displayName + " = " + str(self.positionForMassFlowSolver) + "\n"
            equationNr = 1
            return resStr, equationNr
        return "", 0

    def exportDivSetting1(self):
        if self.isTempering:
            constants = 1
            line = "T_set_" + self.displayName + "=50\n"
            return line, constants
        return "", 0

    def exportDivSetting2(self, nUnit):
        if self.isTempering:
            lines = ""
            nUnit = nUnit + 1
            lines += f"UNIT {nUnit} TYPE 811 ! Passive Divider for heating \n"
            lines += "PARAMETERS 1" + "\n"
            lines += "5 !Nb.of iterations before fixing the value \n"
            lines += "INPUTS 4 \n"

            if (
                    self.outputs[0].pos().y() == self.inputs[0].pos().y()
                    or self.outputs[0].pos().x() == self.inputs[0].pos().x()
            ):
                first = self.outputs[0]
                second = self.outputs[1]

            lines += "T" + first.connectionList[0].displayName + "\n"
            lines += "T" + second.connectionList[0].displayName + "\n"
            lines += "Mfr" + self.inputs[0].connectionList[0].displayName + "\n"

            lines += "T_set_" + self.displayName + "\n"
            lines += "*** INITIAL INPUT VALUES" + "\n"
            lines += "35.0 21.0 800.0 T_set_" + self.displayName + "\n"

            lines += "EQUATIONS 1\n"
            lines += f"xFrac{self.displayName} =  1.-[{nUnit},5] \n\n"

            return lines, nUnit
        return "", nUnit

    def getInternalPiping(self) -> _mfs.InternalPiping:
        teePiece, modelPortItemsToGraphicalPortItem = self._getModelAndMapping()

        return _mfs.InternalPiping([teePiece], modelPortItemsToGraphicalPortItem)

    def _getModelAndMapping(self):
        inputPort = _mfn.PortItem("input", _mfn.PortItemType.INPUT)
        output1 = _mfn.PortItem("straightOutput", _mfn.PortItemType.OUTPUT)
        output2 = _mfn.PortItem("orthogonalOutput", _mfn.PortItemType.OUTPUT)
        teePiece = _mfn.Diverter(self.displayName, self.trnsysId, inputPort, output1, output2)
        modelPortItemsToGraphicalPortItem = {inputPort: self.inputs[0], output1: self.outputs[0],
                                             output2: self.outputs[1]}
        return teePiece, modelPortItemsToGraphicalPortItem

    def exportPipeAndTeeTypesForTemp(self, startingUnit):
        if self.isVisible():
            lines = ""
            unitNumber = startingUnit
            tNr = 929  # Temperature calculation from a tee-piece

            unitText = ""
            ambientT = 20

            equationConstant = 1

            unitText += "UNIT " + str(unitNumber) + " TYPE " + str(tNr) + "\n"
            unitText += "!" + self.displayName + "\n"
            unitText += "PARAMETERS 0\n"
            unitText += "INPUTS 6\n"

            openLoops, _ = self._getOpenLoopsAndNodeToIndices()
            assert len(openLoops) == 1
            openLoop = openLoops[0]

            assert len(openLoop.realNodes) == 1
            realNode = openLoop.realNodes[0]

            outputVariables = realNode.getOutputVariables()
            for outputVariable in outputVariables:
                if not outputVariable:
                    continue

                unitText += outputVariable.name + "\n"

            unitText += f"T{self.inputs[0].connectionList[0].displayName}\n"
            unitText += f"T{self.outputs[0].connectionList[0].displayName}\n"
            unitText += f"T{self.outputs[1].connectionList[0].displayName}\n"

            unitText += "***Initial values\n"
            unitText += 3 * "0 " + 3 * (str(ambientT) + " ") + "\n"

            unitText += "EQUATIONS 1\n"
            unitText += "T" + self.displayName + "= [" + str(unitNumber) + "," + str(equationConstant) + "]\n"

            unitNumber += 1
            lines += unitText + "\n"

            return lines, unitNumber
        return "", startingUnit


@_dc.dataclass
class TVentilModel(_ser.UpgradableJsonSchemaMixin):  # pylint: disable=too-many-instance-attributes
    BlockName: str  # pylint: disable = invalid-name
    BlockDisplayName: str  # pylint: disable = invalid-name
    blockPosition: _tp.Tuple[float, float]
    Id: int  # pylint: disable = invalid-name
    trnsysId: int
    portsIdsIn: _tp.List[int]
    portsIdsOut: _tp.List[int]
    flippedH: bool
    flippedV: bool
    rotationN: int

    @classmethod
    def from_dict(
            cls,  # pylint: disable = duplicate-code
            data: _dcj.JsonDict,  # pylint: disable = duplicate-code
            validate=True,
            validate_enums: bool = True,
    ) -> "TVentilModel":
        tVentilModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(TVentilModel, tVentilModel)

    def to_dict(
            self,
            omit_none: bool = True,
            validate: bool = False,
            validate_enums: bool = True,  # pylint: disable=duplicate-code
    ) -> _dcj.JsonDict:  # pylint: disable = duplicate-code
        data = super().to_dict(omit_none, validate, validate_enums)
        data[".__BlockDict__"] = True
        return data

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixin]:
        return _bim.BlockItemModel

    @classmethod
    def upgrade(cls, superseded: _bim.BlockItemModel) -> "TVentilModel":  # type: ignore[override]
        assert len(superseded.portsIdsIn) == 2
        assert len(superseded.portsIdsOut) == 1

        inputPortIds = [superseded.portsIdsOut[0]]
        outputPortIds = [superseded.portsIdsIn[0], superseded.portsIdsIn[1]]

        return TVentilModel(
            superseded.BlockName,
            superseded.BlockDisplayName,
            superseded.blockPosition,
            superseded.Id,
            superseded.trnsysId,
            inputPortIds,
            outputPortIds,
            superseded.flippedH,
            superseded.flippedV,
            superseded.rotationN,
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('3fff9a8a-d40e-42e2-824d-c015116d0a1d')
